"""Plotting functions.

Matthew Alger
The Australian National University
2016
"""

import astropy.io.fits
import astropy.wcs
import matplotlib.pyplot as plt
from matplotlib.pyplot import cm
import numpy
import sklearn.metrics

from .config import config

ARCMIN = 1 / 60  # deg
FITS_CONVENTION = 1
IMAGE_DIAMETER = config['surveys']['atlas']['fits_width']

wcs = None

def _init_wcs():
    """Initialise the ATLAS image WCS. Sets global variable wcs."""
    with astropy.io.fits.open(config['data_sources']['atlas_image'],
                              ignore_blank=True) as atlas_image:
        global wcs
        wcs = astropy.wcs.WCS(atlas_image[0].header).dropaxis(3).dropaxis(2)


def ra_dec_to_pixels(subject_coords, coords):
    if wcs is None:
        _init_wcs()

    offset, = wcs.all_world2pix([subject_coords], FITS_CONVENTION)
    # The coords are of the middle of the subject.
    coords = wcs.all_world2pix(coords, FITS_CONVENTION)
    coords -= offset
    
    coords[:, 0] /= config['surveys']['atlas']['mosaic_scale_x'] * 424 / 200
    coords[:, 1] /= config['surveys']['atlas']['mosaic_scale_y'] * 424 / 200
    
    coords += [40, 40]
    
    return coords


def plot_points_on_background(points, background, noise=False, base_size=200):
    plt.imshow(background, cmap='gray')

    colours = cm.rainbow(numpy.linspace(0, 1, len(points)))
    for colour, (x, y) in zip(colours, points):
        if noise:
            x += numpy.random.normal(scale=0.5)
            y += numpy.random.normal(scale=0.5)
        plt.scatter(x, y, marker='x', c=colour, s=base_size)
    plt.axis('off')
    plt.xlim((0, background.shape[0]))
    plt.ylim((0, background.shape[1]))


def plot_classifications(atlas_vector, ir_matrix, labels, base_size=200,
                         noise=False):
    image = atlas_vector[2 : 2 + IMAGE_DIAMETER ** 2].reshape(
            (IMAGE_DIAMETER, IMAGE_DIAMETER))[60:140, 60:140]
    radio_coords = atlas_vector[:2]

    nearby = atlas_vector[2 + IMAGE_DIAMETER ** 2:] < ARCMIN
    nearby_labels = labels[nearby]

    ir_coords_ = ir_matrix[nearby, :2][nearby_labels.astype(bool)]
    ir_coords_ = ra_dec_to_pixels(radio_coords, ir_coords_)

    ir_coords = []
    for label, coords in zip(nearby_labels.nonzero()[0], ir_coords_):
        label = nearby_labels[label]
        ir_coords.extend([coords] * label)  # For label multiplicity.
    
    plot_points_on_background(ir_coords, image, base_size=base_size,
                              noise=noise)


def plot_classifications_row(atlas_vector, ir_matrix, classifier_labels,
                             rgz_labels, norris_labels, base_size=200,
                             noise=False):
    plt.subplot(1, 3, 1)
    plt.title('Classifier')
    plot_classifications(atlas_vector, ir_matrix, classifier_labels,
                         base_size=base_size, noise=noise)
    
    plt.subplot(1, 3, 2)
    plt.title('RGZ')
    plot_classifications(atlas_vector, ir_matrix, rgz_labels,
                         base_size=base_size, noise=noise)
    
    plt.subplot(1, 3, 3)
    plt.title('Norris')
    plot_classifications(atlas_vector, ir_matrix, norris_labels,
                         base_size=base_size, noise=noise)


def vertical_scatter(xs, ys, style='bx', rotation='horizontal'):
    """Plots a vertical scatter plot.

    xs: List of x labels.
    ys: List of lists of points to scatter vertically.
    style: Plots point style. Default 'bx'.
    rotation: x label rotation. Default 'horizontal'.
    """
    for x in range(len(xs)):
        plt.plot([x] * len(ys[x]), ys[x], style)
    plt.xticks(range(len(xs)), xs, rotation=rotation)


def vertical_scatter_ba(results, targets, ylim=(0.7, 1.0), **kwargs):
    """Plot a vertical scatter plot of balanced accuracies.

    results: Results object.
    targets: Target labels.
    ylim: (lower, upper) y axis.
    kwargs: Keyword arguments passed to vertical_scatter.
    """
    xs = sorted(results.methods, key=results.methods.get)
    ys = []
    for method in xs:
        y = []
        for split in range(results.n_splits):
            mask = results.get_mask(method, split)
            split_results = results[method, split][mask].round()
            split_targets = targets[mask]
            if len(split_results) == 0:
                continue
            # Calculate balanced accuracy.
            cm = sklearn.metrics.confusion_matrix(split_targets, split_results)
            tp = cm[1, 1]
            n, p = cm.sum(axis=1)
            tn = cm[0, 0]
            ba = (tp / p + tn / n) / 2
            y.append(ba)
        ys.append(y)

    vertical_scatter(xs, ys, **kwargs)
    plt.ylim(ylim)
    plt.xlim((-0.5, len(xs) - 0.5))  # Adds a little buffer.
    plt.ylabel('Balanced accuracy (%)')



"""Loads the configuration file.

Matthew Alger
The Australian National University
2016
"""

import json
import os.path

import numpy

SCRIPT_PATH = os.path.dirname(os.path.realpath(__file__))

# TODO(MatthewJA): Don't hardcode this.
CONFIG_PATH = os.path.join(SCRIPT_PATH, '../crowdastro.json')

with open(CONFIG_PATH) as config_file:
    config = json.load(config_file)

# Normalise paths.
# TODO(MatthewJA): Also don't hardcode these. They should be relative to the
# JSON(?).
config['data_sources']['atlas_catalogue'] = os.path.normpath(
        os.path.join(SCRIPT_PATH, '..',
                     config['data_sources']['atlas_catalogue']))
config['data_sources']['swire_catalogue'] = os.path.normpath(
        os.path.join(SCRIPT_PATH, '..',
                     config['data_sources']['swire_catalogue']))
config['data_sources']['atlas_image'] = os.path.normpath(
        os.path.join(SCRIPT_PATH, '..',
                     config['data_sources']['atlas_image']))
config['data_sources']['norris_coords'] = os.path.normpath(
        os.path.join(SCRIPT_PATH, '..',
                     config['data_sources']['norris_coords']))

# Generate some helper configuration info.
config['surveys']['atlas']['click_to_fits_x'] = (
    config['surveys']['atlas']['fits_width'] /
    config['surveys']['atlas']['click_width'])
config['surveys']['atlas']['click_to_fits_y'] = (
    config['surveys']['atlas']['fits_height'] /
    config['surveys']['atlas']['click_height'])
config['surveys']['atlas']['click_to_fits'] = numpy.array(
    [config['surveys']['atlas']['click_to_fits_x'],
     config['surveys']['atlas']['click_to_fits_y']],
    dtype=float)
config['surveys']['atlas']['web_to_click_x'] = (
    config['surveys']['atlas']['click_width'] /
    config['surveys']['atlas']['web_width'])
config['surveys']['atlas']['web_to_click_y'] = (
    config['surveys']['atlas']['click_height'] /
    config['surveys']['atlas']['web_height'])
config['surveys']['atlas']['web_to_click'] = numpy.array(
    [config['surveys']['atlas']['web_to_click_x'],
     config['surveys']['atlas']['web_to_click_y']],
    dtype=float)

# Wrapper function to neatly retrieve values.
def get(*args, **kwargs):
    return config.get(*args, **kwargs)



"""Generates sets of testing indices for the galaxy classification task.

Matthew Alger
The Australian National University
2016
"""

import argparse

import h5py
import numpy

from .config import config


ATLAS_WIDTH = config['surveys']['atlas']['fits_width']
ATLAS_HEIGHT = config['surveys']['atlas']['fits_height']
ATLAS_SIZE = ATLAS_WIDTH * ATLAS_HEIGHT


def get_nearby_galaxies(atlas_vector, radius=1/60):
    """Gets all nearby galaxy IDs.

    atlas_vector: crowdastro ATLAS vector.
    radius: Radius to call a galaxy "nearby", in degrees. Default 1 arcmin.
    -> list of ints
    """
    return (atlas_vector[2 + ATLAS_SIZE:] <= radius).nonzero()[0]


def main(f_h5, n, p):
    ir_survey = f_h5.attrs['ir_survey']
    fan_labels = f_h5['/{}/cdfs/fan_labels'.format(ir_survey)].value
    norris_labels = f_h5['/{}/cdfs/norris_labels'.format(ir_survey)].value

    candidates = []
    for atlas_id, atlas in enumerate(f_h5['/atlas/cdfs/numeric']):
        # Get all nearby galaxies.
        nearby = get_nearby_galaxies(atlas)
        # If Norris and Fan agree on all nearby galaxies, then add this ATLAS
        # object to a set of candidates.
        if all(fan_labels[n] == norris_labels[n] for n in nearby):
            candidates.append(atlas_id)

    n_atlas = f_h5['/atlas/cdfs/numeric'].shape[0]
    test_sets = []
    for i in range(n):
        # Select at random, without replacement, candidate ATLAS objects.
        candidates_ = candidates[:]
        numpy.random.shuffle(candidates_)
        atlas_test_set = candidates_[:int(n_atlas * p)]

        # Get all nearby galaxies and add all nearby galaxies to the test set.
        test_set = []
        for atlas_id in atlas_test_set:
            nearby = get_nearby_galaxies(f_h5['/atlas/cdfs/numeric'][atlas_id])
            test_set.extend(nearby)

        test_sets.append(list(set(test_set)))

    # Because the test sets are galaxy-based but the drawing was radio-based,
    # we may have unequal length lists, so we'll crop them.
    min_length = min(len(test_set) for test_set in test_sets)
    test_sets_ = []
    for test_set in test_sets:
        # Have to reshuffle so we don't bias against later IDs.
        numpy.random.shuffle(test_set)
        test_sets_.append(sorted(test_set[:min_length]))

    test_sets = numpy.array(test_sets_)
    f_h5.create_dataset('/{}/cdfs/test_sets'.format(ir_survey),
                        data=test_sets)


def _populate_parser(parser):
    parser.description = 'Generates sets of testing indices for the galaxy ' \
                         'classification task.'
    parser.add_argument('--h5', default='data/crowdastro.h5',
                        help='Crowdastro HDF5 file')
    parser.add_argument('--n', default=5, type=int,
                        help='Number of test sets')
    parser.add_argument('--p', default=0.5, type=float,
                        help='Percentage size of test sets')


def _main(args):
    with h5py.File(args.h5, 'r+') as f_h5:
        main(f_h5, args.n, args.p)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Generates CNN output training data.

Matthew Alger
The Australian National University
2016
"""

import argparse
import csv
import logging

import h5py
import numpy

from .config import config

PATCH_RADIUS = config['patch_radius']  # px
ARCMIN = 1 / 60  # deg


def generate(training_h5, cnn_model_json, cnn_weights_path):
    """Generates CNN output training data.

    training_h5: HDF5 file with training data.
    cnn_model_path: JSON model file.
    cnn_weights_path: Path to CNN weights HDF5 file.
    """
    # Have to import Keras here because it dumps to stdout...
    import keras
    n_static = 6 if training_h5.attrs['ir_survey'] == 'swire' else 7
    cnn = keras.models.model_from_json(cnn_model_json.read())
    cnn.load_weights(cnn_weights_path)
    cnn.compile(loss='binary_crossentropy', optimizer='adadelta')
    get_convolutional_features_ = keras.backend.function(
            [cnn.layers[0].input], [cnn.layers[5].output])
    get_convolutional_features = (lambda p:
            get_convolutional_features_([p])[0].reshape((p.shape[0], -1)))

    images = training_h5['features'][:, n_static:].reshape(
            (-1, 1, PATCH_RADIUS * 2, PATCH_RADIUS * 2))

    test_out = get_convolutional_features(images[:1, :, :, :])

    if '_features' in training_h5:
        del training_h5['_features']

    out = training_h5.create_dataset('_features', dtype=float,
            shape=(len(images), n_static + test_out.shape[1]))

    # Copy the static features across. We'll fill in the rest with the CNN.
    out[:, :n_static] = training_h5['features'][:, :n_static]

    batch_size = 1000
    for i in range(0, len(images), batch_size):
        batch = images[i : i + batch_size]
        out[i : i + batch_size, n_static:] = get_convolutional_features(batch)

    # Clean up - delete the original features and rename our new features.
    del training_h5['features']
    training_h5['features'] = training_h5['_features']
    del training_h5['_features']


def _populate_parser(parser):
    parser.description = 'Generates convolutional neural network outputs for ' \
                         'training data.'
    parser.add_argument('--training', default='data/training.h5',
                        help='HDF5 training file')
    parser.add_argument('--model', default='data/model.json',
                        help='JSON CNN model')
    parser.add_argument('--weights', default='data/weights.h5',
                        help='HDF5 CNN weights')


def _main(args):
    with h5py.File(args.training, 'r+') as training_h5:
        with open(args.model, 'r') as cnn_model_json:
            generate(training_h5, cnn_model_json, args.weights)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Command-line interface for crowdastro.

Matthew Alger
The Australian National University
2016
"""

import argparse
import logging
import sys

from . import __description__
from . import compile_cnn
from . import consensuses
from . import generate_cnn_outputs
from . import generate_dataset
from . import generate_test_sets
from . import generate_training_data
from . import import_data
from . import repack_h5
from . import test
from . import train
from . import train_cnn

def main():
    parser = argparse.ArgumentParser(description=__description__)
    parser.add_argument('--verbose', '--v', '-v', action='store_true',
            help='verbose output')
    subparsers = parser.add_subparsers(dest='subcommand')

    parser_compile_cnn = subparsers.add_parser('compile_cnn',
            help='compile a convolutional neural network')
    compile_cnn._populate_parser(parser_compile_cnn)

    parser_consensuses = subparsers.add_parser('consensuses',
            help='generate Radio Galaxy Zoo consensus classifications')
    consensuses._populate_parser(parser_consensuses)

    parser_generate_cnn_outputs = subparsers.add_parser('generate_cnn_outputs',
            help='generate convolutional neural network training outputs')
    generate_cnn_outputs._populate_parser(parser_generate_cnn_outputs)

    parser_generate_dataset = subparsers.add_parser('generate_dataset',
            help='generate crowdastro dataset')
    generate_dataset._populate_parser(parser_generate_dataset)

    parser_generate_test_sets = subparsers.add_parser('generate_test_sets',
            help='generate crowdastro galaxy test sets')
    generate_test_sets._populate_parser(parser_generate_test_sets)

    parser_generate_training_data = subparsers.add_parser(
            'generate_training_data',
            help='generate crowdastro galaxy test sets')
    generate_training_data._populate_parser(parser_generate_training_data)

    parser_import_data = subparsers.add_parser('import_data',
            help='import data into crowdastro')
    import_data._populate_parser(parser_import_data)

    parser_repack_h5 = subparsers.add_parser('repack_h5',
            help='repacks an HDF5 file')
    repack_h5._populate_parser(parser_repack_h5)

    parser_test = subparsers.add_parser('test',
            help='tests classifiers')
    test._populate_parser(parser_test)

    parser_train = subparsers.add_parser('train',
            help='trains classifiers')
    train._populate_parser(parser_train)

    parser_train_cnn = subparsers.add_parser('train_cnn',
            help='trains a convolutional neural network')
    train_cnn._populate_parser(parser_train_cnn)

    # http://stackoverflow.com/a/11287731/1105803
    if len(sys.argv) < 2:
        sys.argv.append('--help')

    args = parser.parse_args()

    logging.captureWarnings(True)
    if args.verbose:
        logging.root.setLevel(logging.DEBUG)

    subcommands = {
        'compile_cnn': compile_cnn._main,
        'consensuses': consensuses._main,
        'generate_cnn_outputs': generate_cnn_outputs._main,
        'generate_dataset': generate_dataset._main,
        'generate_test_sets': generate_test_sets._main,
        'generate_training_data': generate_training_data._main,
        'import_data': import_data._main,
        'repack_h5': repack_h5._main,
        'test': test._main,
        'train': train._main,
        'train_cnn': train_cnn._main,
    }

    subcommands[args.subcommand](args)


if __name__ == '__main__':
    main()



"""Trains classifiers.

Matthew Alger
The Australian National University
2016
"""

import argparse
import logging

import h5py
import numpy
import sklearn.externals.joblib
import sklearn.ensemble
import sklearn.kernel_approximation
import sklearn.linear_model
import sklearn.pipeline
import sklearn.preprocessing
import sklearn.svm

from .config import config

PATCH_RADIUS = config['patch_radius']
PATCH_DIAMETER = PATCH_RADIUS * 2

def train(training_h5, classifier_out_path, astro_transformer_out_path,
          image_transformer_out_path, classifier='lr', use_astro=True,
          use_cnn=True, n_jobs=-1):
    """Trains logistic regression.

    training_h5: Training HDF5 file.
    classifier_out_path: Output classifier path.
    astro_transformer_out_path: Output astro transformer path.
    image_transformer_out_path: Output image transformer path.
    classifier: Classifier to use in {'lr', 'rf'}. Default 'lr'.
    use_astro: Use astronomical features. Default True.
    use_cnn: Use CNN features. Default True.
    n_jobs: Number of cores to use. Default all.
    -> classifier
    """
    if not any([use_astro, use_cnn]):
        raise ValueError('Must have features.')
    n_static = 6 if training_h5.attrs['ir_survey'] == 'swire' else 5

    train_indices = training_h5['is_ir_train'].value
    outputs = training_h5['labels'].value[train_indices]
    n = len(outputs)

    astro_inputs = numpy.minimum(
            training_h5['features'][train_indices, :n_static], 1500)
    image_inputs = training_h5['features'].value[train_indices, n_static:]

    astro_transformer = sklearn.pipeline.Pipeline([
            ('normalise', sklearn.preprocessing.Normalizer()),
            ('scale', sklearn.preprocessing.StandardScaler()),
    ])
    image_transformer = sklearn.pipeline.Pipeline([
            ('normalise', sklearn.preprocessing.Normalizer()),
    ])

    features = []
    if use_astro:
        features.append(astro_transformer.fit_transform(astro_inputs))
    if use_cnn:
        features.append(image_transformer.fit_transform(image_inputs))
    inputs = numpy.hstack(features)

    if classifier == 'lr':
        classifier = sklearn.linear_model.LogisticRegression(
                class_weight='balanced', n_jobs=n_jobs)
    elif classifier == 'rf':
        classifier = sklearn.ensemble.RandomForestClassifier(
                class_weight='balanced', n_jobs=n_jobs)
    elif classifier == 'svm':
        classifier = sklearn.svm.SVC(class_weight='balanced', probability=True)
    elif classifier == 'klr':
        # A bit of a hack to reduce the size of the data so we can guess a gamma
        # value. Otherwise we run out of memory!
        # TODO(MatthewJA): This will break for bigger data sets.
        input_sample, _ = sklearn.cross_validation.train_test_split(
                inputs, train_size=0.2, stratify=outputs)
        dists = sklearn.metrics.pairwise_distances(input_sample, n_jobs=n_jobs)
        gamma = 1 / numpy.percentile(dists ** 2, 75)
        sampler = sklearn.kernel_approximation.RBFSampler(gamma=gamma,
                n_components=1000)
        lr = sklearn.linear_model.LogisticRegression(
                class_weight='balanced', n_jobs=n_jobs)
        classifier = sklearn.pipeline.Pipeline([
                ("feature_map", sampler),
                ("lr", lr)])
    else:
        raise ValueError('Unknown classifier: {}'.format(classifier))

    classifier.fit(inputs, outputs)

    sklearn.externals.joblib.dump(classifier, classifier_out_path)
    sklearn.externals.joblib.dump(astro_transformer, astro_transformer_out_path)
    sklearn.externals.joblib.dump(image_transformer, image_transformer_out_path)

    return classifier, astro_transformer, image_transformer


def _populate_parser(parser):
    parser.description = 'Trains classifiers.'
    parser.add_argument('--training', default='data/training.h5',
                        help='HDF5 training data file')
    parser.add_argument('--c_out', default='data/classifier.pkl',
                        help='classifier output file')
    parser.add_argument('--at_out', default='data/astro_transformer.pkl',
                        help='astro_transformer output file')
    parser.add_argument('--it_out', default='data/image_transformer.pkl',
                        help='image_transformer output file')
    parser.add_argument('--classifier',
                        choices={'lr', 'rf', 'svm', 'klr'},
                        default='lr', help='which classifier to train')
    parser.add_argument('--no_astro', action='store_false', default=True,
                        help='ignore astro features')
    parser.add_argument('--no_cnn', action='store_false', default=True,
                        help='ignore CNN features')
    parser.add_argument('--n_jobs', default=-1, type=int,
                        help='number of cores to use')


def _main(args):
    with h5py.File(args.training, 'r') as training_h5:
        train(training_h5, args.c_out, args.at_out, args.it_out,
              classifier=args.classifier, use_astro=args.no_astro,
              use_cnn=args.no_cnn)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Machine learning using crowdsourced labels on the Radio Galaxy Zoo.

MIT-licensed. See license.txt.

Matthew Alger & Cheng Soon Ong
The Australian National University
2016
"""

__version__ = '0.0.1a1'
__description__ = 'Automated cross-identification of radio objects and host ' \
                  'galaxies using crowdsourced labels from the Radio Galaxy ' \
                  'Zoo.'



"""Imports and standardises data into crowdastro.

Matthew Alger
The Australian National University
2016
"""

import argparse
import csv
import logging
import warnings

from astropy.coordinates import SkyCoord
import astropy.io.fits
import astropy.utils.exceptions
import astropy.wcs
import h5py
import numpy
import scipy.spatial.distance
import sklearn.neighbors

from .config import config
from .exceptions import CatalogueError

VERSION = '0.5.1'  # Data version, not module version!
MAX_RADIO_SIGNATURE_LENGTH = 50  # max number of components * individual
                                 # component signature size.
MAX_NAME_LENGTH = 50  # b
MAX_ZOONIVERSE_ID_LENGTH = 20  # b
PATCH_RADIUS = config['patch_radius']  # px
ARCMIN = 1 / 60  # deg
CANDIDATE_RADIUS = ARCMIN  # deg
FITS_CONVENTION = 1

def prep_h5(f_h5, ir_survey):
    """Creates hierarchy in HDF5 file."""
    f_h5.create_group('/atlas/cdfs')
    f_h5.create_group('/{}/cdfs'.format(ir_survey))
    f_h5.attrs['version'] = VERSION
    f_h5.attrs['ir_survey'] = ir_survey


def import_atlas(f_h5, test=False):
    """Imports the ATLAS dataset into crowdastro, as well as associated SWIRE.

    f_h5: An HDF5 file.
    test: Flag to run on only 10 subjects. Default False.
    """
    from . import rgz_data as data

    # Fetch groups from HDF5.
    cdfs = f_h5['/atlas/cdfs']

    # First pass, I'll find coords, names, and Zooniverse IDs, as well as how
    # many data points there are.

    coords = []
    names = []
    zooniverse_ids = []

    # We need the ATLAS name, but we can only get it by going through the
    # ATLAS catalogue and finding the nearest component.
    # https://github.com/chengsoonong/crowdastro/issues/63
    # Fortunately, @jbanfield has already done this, so we can just load
    # that CSV and match the names.
    # TODO(MatthewJA): This matches the ATLAS component ID, but maybe we should\
    # be using the name instead.
    rgz_to_atlas = {}
    with open(config['data_sources']['rgz_to_atlas']) as f:
        reader = csv.DictReader(f)
        for row in reader:
            rgz_to_atlas[row['ID_RGZ']] = row['ID']

    all_subjects = data.get_all_subjects(survey='atlas', field='cdfs')
    if test:
        all_subjects = all_subjects.limit(10)

    for subject in all_subjects:
        ra, dec = subject['coords']
        zooniverse_id = subject['zooniverse_id']

        rgz_source_id = subject['metadata']['source']
        if rgz_source_id not in rgz_to_atlas:
            logging.debug('Skipping %s; no matching ATLAS component.',
                          zooniverse_id)
            continue
        name = rgz_to_atlas[rgz_source_id]

        # Store the results.
        coords.append((ra, dec))
        names.append(name)
        zooniverse_ids.append(zooniverse_id)

    n_cdfs = len(names)

    # Sort the data by Zooniverse ID.
    coords_to_zooniverse_ids = dict(zip(coords, zooniverse_ids))
    names_to_zooniverse_ids = dict(zip(names, zooniverse_ids))

    coords.sort(key=coords_to_zooniverse_ids.get)
    names.sort(key=names_to_zooniverse_ids.get)
    zooniverse_ids.sort()

    # Begin to store the data. We will have two tables: one for numeric data,
    # and one for strings. We will have to preallocate the numeric table so that
    # we aren't storing huge amounts of image data in memory.

    # Strings.
    dtype = [('zooniverse_id', '<S{}'.format(MAX_ZOONIVERSE_ID_LENGTH)),
             ('name', '<S{}'.format(MAX_NAME_LENGTH))]
    string_data = numpy.array(list(zip(zooniverse_ids, names)), dtype=dtype)
    cdfs.create_dataset('string', data=string_data, dtype=dtype)

    # Numeric.
    image_size = (config['surveys']['atlas']['fits_width'] *
                    config['surveys']['atlas']['fits_height'])
    # RA, DEC, radio, (distance to SWIRE object added later)
    dim = (n_cdfs, 1 + 1 + image_size)
    numeric = cdfs.create_dataset('_numeric', shape=dim, dtype='float32')

    # Load image patches and store numeric data.
    with astropy.io.fits.open(config['data_sources']['atlas_image'],
                              ignore_blank=True) as atlas_image:
        wcs = astropy.wcs.WCS(atlas_image[0].header).dropaxis(3).dropaxis(2)
        pix_coords = wcs.all_world2pix(coords, FITS_CONVENTION)
        assert pix_coords.shape[1] == 2
        logging.debug('Fetching %d ATLAS images.', len(pix_coords))
        for index, (x, y) in enumerate(pix_coords):
            radio = atlas_image[0].data[0, 0,  # stokes, freq
                    int(y) - config['surveys']['atlas']['fits_height'] // 2 :
                    int(y) + config['surveys']['atlas']['fits_height'] // 2 ,
                    int(x) - config['surveys']['atlas']['fits_width'] // 2 :
                    int(x) + config['surveys']['atlas']['fits_width'] // 2 ]
            numeric[index, 0] = coords[index][0]
            numeric[index, 1] = coords[index][1]
            numeric[index, 2 : 2 + image_size] = radio.reshape(-1)

    logging.debug('ATLAS imported.')

    # TODO(MatthewJA): Partition into training/testing sets, ideally using
    # expert or gold standard classifications.


def remove_nulls(n):
    """Swaps nulls with zeros."""
    if n == 'null':
        return 0

    return n


def import_swire(f_h5):
    """Imports the SWIRE dataset into crowdastro.

    f_h5: An HDF5 file.
    """
    names = []
    rows = []
    logging.debug('Reading SWIRE catalogue.')
    with open(config['data_sources']['swire_catalogue']) as f_tbl:
        # This isn't a valid ASCII table, so Astropy can't handle it. This means
        # we have to parse it manually.
        for _ in range(5):  # Skip the first five lines.
            next(f_tbl)

        # Get the column names.
        columns = [c.strip() for c in next(f_tbl).strip().split('|')][1:-1]
        assert len(columns) == 156

        for _ in range(3):  # Skip the next three lines.
            next(f_tbl)

        for row in f_tbl:
            row = row.strip().split()
            assert len(row) == 156
            row = dict(zip(columns, row))
            name = row['object']
            ra = float(row['ra'])
            dec = float(row['dec'])
            flux_ap2_36 = float(remove_nulls(row['flux_ap2_36']))
            flux_ap2_45 = float(remove_nulls(row['flux_ap2_45']))
            flux_ap2_58 = float(remove_nulls(row['flux_ap2_58']))
            flux_ap2_80 = float(remove_nulls(row['flux_ap2_80']))
            flux_ap2_24 = float(remove_nulls(row['flux_ap2_24']))
            stell_36 = float(remove_nulls(row['stell_36']))
            # Extra -1 is so we can store nearest distance later.
            rows.append((ra, dec, flux_ap2_36, flux_ap2_45, flux_ap2_58,
                         flux_ap2_80, flux_ap2_24, stell_36, -1))
            names.append(name)

    logging.debug('Found %d SWIRE objects.', len(names))

    # Sort by name.
    rows_to_names = dict(zip(rows, names))
    rows.sort(key=rows_to_names.get)
    names.sort()

    names = numpy.array(names, dtype='<S{}'.format(MAX_NAME_LENGTH))
    rows = numpy.array(rows)

    # Filter on distance - only include image data for SWIRE objects within a
    # given radius of an ATLAS object. Otherwise, there's way too much data to
    # store.
    swire_positions = rows[:, :2]
    atlas_positions = f_h5['/atlas/cdfs/_numeric'][:, :2]
    logging.debug('Computing SWIRE k-d tree.')
    swire_tree = sklearn.neighbors.KDTree(swire_positions, metric='euclidean')
    indices = numpy.concatenate(
            swire_tree.query_radius(atlas_positions, CANDIDATE_RADIUS))
    indices = numpy.unique(indices)

    logging.debug('Found %d SWIRE objects near ATLAS objects.', len(indices))

    names = names[indices]
    rows = rows[indices]
    swire_positions = swire_positions[indices]

    # Get distances.
    logging.debug('Finding ATLAS-SWIRE object distances.')
    distances = scipy.spatial.distance.cdist(atlas_positions, swire_positions,
                                             'euclidean')
    assert distances.shape[0] == atlas_positions.shape[0]
    assert distances.shape[1] == swire_positions.shape[0]
    logging.debug('Done finding distances.')

    # Write numeric data to HDF5.
    rows[:, 8] = distances.min(axis=0)
    atlas_numeric = f_h5['/atlas/cdfs/_numeric']
    f_h5['/atlas/cdfs'].create_dataset('numeric', dtype='float32',
            shape=(atlas_numeric.shape[0],
                   atlas_numeric.shape[1] + len(indices)))
    f_h5['/atlas/cdfs/numeric'][:, :atlas_numeric.shape[1]] = atlas_numeric
    f_h5['/atlas/cdfs/numeric'][:, atlas_numeric.shape[1]:] = distances

    del f_h5['/atlas/cdfs/_numeric']

    image_size = (PATCH_RADIUS * 2) ** 2
    dim = (rows.shape[0], rows.shape[1] + image_size)
    numeric = f_h5['/swire/cdfs'].create_dataset('numeric', shape=dim,
                                                 dtype='float32')
    numeric[:, :rows.shape[1]] = rows
    f_h5['/swire/cdfs'].create_dataset('string', data=names)

    # Load and store radio images.
    logging.debug('Importing radio patches.')
    with astropy.io.fits.open(config['data_sources']['atlas_image'],
                              ignore_blank=True) as atlas_image:
        wcs = astropy.wcs.WCS(atlas_image[0].header).dropaxis(3).dropaxis(2)
        pix_coords = wcs.all_world2pix(swire_positions, FITS_CONVENTION)
        assert pix_coords.shape[1] == 2
        assert pix_coords.shape[0] == len(indices)
        logging.debug('Fetching %d ATLAS patches.', len(indices))

        for index, (x, y) in enumerate(pix_coords):
            radio = atlas_image[0].data[0, 0,  # stokes, freq
                    int(y) - PATCH_RADIUS :
                    int(y) + PATCH_RADIUS ,
                    int(x) - PATCH_RADIUS :
                    int(x) + PATCH_RADIUS ]
            numeric[index, -image_size:] = radio.reshape(-1)


def import_wise(f_h5):
    """Imports the WISE dataset into crowdastro.

    f_h5: An HDF5 file.
    """
    names = []
    rows = []
    logging.debug('Reading WISE catalogue.')
    with open(config['data_sources']['wise_catalogue']) as f_tbl:
        # This isn't a valid ASCII table, so Astropy can't handle it. This means
        # we have to parse it manually.
        for _ in range(105):  # Skip the first 105 lines.
            next(f_tbl)

        # Get the column names.
        columns = [c.strip() for c in next(f_tbl).strip().split('|')][1:-1]
        assert len(columns) == 45

        for _ in range(3):  # Skip the next three lines.
            next(f_tbl)

        for row in f_tbl:
            row = row.strip().split()
            assert len(row) == 45
            row = dict(zip(columns, row))
            name = row['designation']
            ra = float(row['ra'])
            dec = float(row['dec'])
            w1mpro = float(remove_nulls(row['w1mpro']))
            w2mpro = float(remove_nulls(row['w2mpro']))
            w3mpro = float(remove_nulls(row['w3mpro']))
            w4mpro = float(remove_nulls(row['w4mpro']))
            # Extra -1 is so we can store nearest distance later.
            rows.append((ra, dec, w1mpro, w2mpro, w3mpro, w4mpro, -1))
            names.append(name)

    logging.debug('Found %d WISE objects.', len(names))

    # Sort by name.
    rows_to_names = dict(zip(rows, names))
    rows.sort(key=rows_to_names.get)
    names.sort()

    names = numpy.array(names, dtype='<S{}'.format(MAX_NAME_LENGTH))
    rows = numpy.array(rows)

    # Filter on distance - only include image data for WISE objects within a
    # given radius of an ATLAS object. Otherwise, there's way too much data to
    # store.
    wise_positions = rows[:, :2]
    atlas_positions = f_h5['/atlas/cdfs/_numeric'][:, :2]
    logging.debug('Computing WISE k-d tree.')
    wise_tree = sklearn.neighbors.KDTree(wise_positions, metric='euclidean')
    indices = numpy.concatenate(
            wise_tree.query_radius(atlas_positions, CANDIDATE_RADIUS))
    indices = numpy.unique(indices)

    logging.debug('Found %d WISE objects near ATLAS objects.', len(indices))

    names = names[indices]
    rows = rows[indices]
    wise_positions = wise_positions[indices]

    # Get distances.
    logging.debug('Finding ATLAS-WISE object distances.')
    distances = scipy.spatial.distance.cdist(atlas_positions, wise_positions,
                                             'euclidean')
    assert distances.shape[0] == atlas_positions.shape[0]
    assert distances.shape[1] == wise_positions.shape[0]
    logging.debug('Done finding distances.')

    # Write numeric data to HDF5.
    rows[:, 6] = distances.min(axis=0)
    atlas_numeric = f_h5['/atlas/cdfs/_numeric']
    f_h5['/atlas/cdfs'].create_dataset('numeric', dtype='float32',
            shape=(atlas_numeric.shape[0],
                   atlas_numeric.shape[1] + len(indices)))
    f_h5['/atlas/cdfs/numeric'][:, :atlas_numeric.shape[1]] = atlas_numeric
    f_h5['/atlas/cdfs/numeric'][:, atlas_numeric.shape[1]:] = distances

    del f_h5['/atlas/cdfs/_numeric']

    image_size = (PATCH_RADIUS * 2) ** 2
    dim = (rows.shape[0], rows.shape[1] + image_size)
    numeric = f_h5['/wise/cdfs'].create_dataset('numeric', shape=dim,
                                                 dtype='float32')
    numeric[:, :rows.shape[1]] = rows
    f_h5['/wise/cdfs'].create_dataset('string', data=names)

    # Load and store radio images.
    logging.debug('Importing radio patches.')
    with astropy.io.fits.open(config['data_sources']['atlas_image'],
                              ignore_blank=True) as atlas_image:
        wcs = astropy.wcs.WCS(atlas_image[0].header).dropaxis(3).dropaxis(2)
        pix_coords = wcs.all_world2pix(wise_positions, FITS_CONVENTION)
        assert pix_coords.shape[1] == 2
        assert pix_coords.shape[0] == len(indices)
        logging.debug('Fetching %d ATLAS patches.', len(indices))

        for index, (x, y) in enumerate(pix_coords):
            radio = atlas_image[0].data[0, 0,  # stokes, freq
                    int(y) - PATCH_RADIUS :
                    int(y) + PATCH_RADIUS ,
                    int(x) - PATCH_RADIUS :
                    int(x) + PATCH_RADIUS ]
            numeric[index, -image_size:] = radio.reshape(-1)


def import_norris(f_h5):
    """Imports the Norris et al. (2006) labels.

    f_h5: crowdastro HDF5 file with WISE or SWIRE already imported.
    """
    ir_survey = f_h5.attrs['ir_survey']
    ir_names = f_h5['/{}/cdfs/string'.format(ir_survey)]
    ir_positions = f_h5['/{}/cdfs/numeric'.format(ir_survey)][:, :2]
    ir_tree = sklearn.neighbors.KDTree(ir_positions)
    with open(config['data_sources']['norris_coords'], 'r') as norris_dat:
        norris_coords = [r.strip().split('|') for r in norris_dat]
    norris_labels = numpy.zeros((ir_positions.shape[0],))
    for ra, dec in norris_coords:
        # Find a neighbour.
        skycoord = SkyCoord(ra=ra, dec=dec, unit=('hourangle', 'deg'))
        ra = skycoord.ra.degree
        dec = skycoord.dec.degree
        ((dist,),), ((ir,),) = ir_tree.query([(ra, dec)])
        if dist < config['surveys'][ir_survey]['distance_cutoff']:
            norris_labels[ir] = 1
    f_h5.create_dataset('/{}/cdfs/norris_labels'.format(ir_survey),
                         data=norris_labels)


def import_fan(f_h5):
    """Imports the Fan et al. (2015) labels.

    f_h5: crowdastro HDF5 file with WISE or SWIRE already imported.
    """
    ir_survey = f_h5.attrs['ir_survey']
    ir_names = f_h5['/{}/cdfs/string'.format(ir_survey)]
    ir_positions = f_h5['/{}/cdfs/numeric'.format(ir_survey)][:, :2]
    ir_tree = sklearn.neighbors.KDTree(ir_positions)
    fan_coords = []
    with open(config['data_sources']['fan_swire'], 'r') as fan_dat:
        for row in csv.DictReader(fan_dat):
            ra_hr = row['swire'][8:10]
            ra_min = row['swire'][10:12]
            ra_sec = row['swire'][12:17]
            dec_sgn = row['swire'][17]
            dec_deg = row['swire'][18:20]
            dec_min = row['swire'][20:22]
            dec_sec = row['swire'][22:26]

            ra = '{} {} {}'.format(ra_hr, ra_min, ra_sec)
            dec = '{}{} {} {}'.format(dec_sgn, dec_deg, dec_min, dec_sec)
            fan_coords.append((ra, dec))

    fan_labels = numpy.zeros((ir_positions.shape[0],))
    for ra, dec in fan_coords:
        # Find a neighbour.
        skycoord = SkyCoord(ra=ra, dec=dec, unit=('hourangle', 'deg'))
        ra = skycoord.ra.degree
        dec = skycoord.dec.degree
        ((dist,),), ((ir,),) = ir_tree.query([(ra, dec)])
        if dist < config['surveys'][ir_survey]['distance_cutoff']:
            fan_labels[ir] = 1
    f_h5.create_dataset('/{}/cdfs/fan_labels'.format(ir_survey),
                         data=fan_labels)


def contains(bbox, point):
    """Checks if point is within bbox.

    bbox: [[x0, x1], [y0, y1]]
    point: [x, y]
    -> bool
    """
    return (bbox[0][0] <= point[0] <= bbox[0][1] and
            bbox[1][0] <= point[1] <= bbox[1][1])


bbox_cache_ = {}  # Should help speed up ATLAS membership checking.


def make_radio_combination_signature(radio_annotation, wcs, atlas_positions,
                                     subject, pix_offset):
    """Generates a unique signature for a radio annotation.

    radio_annotation: 'radio' dictionary from a classification.
    wcs: World coordinate system associated with the ATLAS image.
    atlas_positions: [[RA, DEC]] NumPy array.
    subject: RGZ subject dict.
    pix_offset: (x, y) pixel position of this radio subject on the ATLAS image.
    -> Something immutable
    """
    from . import rgz_data as data
    # TODO(MatthewJA): This only works on ATLAS. Generalise.
    # My choice of immutable object will be stringified crowdastro ATLAS
    # indices.
    zooniverse_id = subject['zooniverse_id']
    subject_fits = data.get_radio_fits(subject)
    subject_wcs = astropy.wcs.WCS(subject_fits.header)

    atlas_ids = []
    x_offset, y_offset = pix_offset
    for c in radio_annotation.values():
        # Note that the x scale is not the same as the IR scale, but the scale
        # factor is included in the annotation, so I have multiplied this out
        # here for consistency.
        scale_width = c.get('scale_width', '')
        scale_height = c.get('scale_height', '')
        if scale_width:
            scale_width = float(scale_width)
        else:
            # Sometimes, there's no scale, so I've included a default scale.
            scale_width = config['surveys']['atlas']['scale_width']

        if scale_height:
            scale_height = float(scale_height)
        else:
            scale_height = config['surveys']['atlas']['scale_height']

        # These numbers are in terms of the PNG images, so I need to multiply by
        # the click-to-fits ratio.
        scale_width *= config['surveys']['atlas']['click_to_fits_x']
        scale_height *= config['surveys']['atlas']['click_to_fits_y']

        subject_bbox = [
            [
                float(c['xmin']) * scale_width,
                float(c['xmax']) * scale_width,
            ],
            [
                float(c['ymin']) * scale_height,
                float(c['ymax']) * scale_height,
            ],
        ]

        # ...and by the mosaic ratio. There's probably double-up here, but this
        # makes more sense.
        scale_width *= config['surveys']['atlas']['mosaic_scale_x']
        scale_height *= config['surveys']['atlas']['mosaic_scale_y']
        # Get the bounding box of the radio source in pixels.
        # Format: [xs, ys]
        bbox = [
            [
                float(c['xmin']) * scale_width,
                float(c['xmax']) * scale_width,
            ],
            [
                float(c['ymin']) * scale_height,
                float(c['ymax']) * scale_height,
            ],
        ]
        assert bbox[0][0] < bbox[0][1]
        assert bbox[1][0] < bbox[1][1]

        # Convert the bounding box into RA/DEC.
        bbox = wcs.all_pix2world(bbox[0] + x_offset, bbox[1] + y_offset,
                                 FITS_CONVENTION)
        subject_bbox = subject_wcs.all_pix2world(subject_bbox[0],
                subject_bbox[1], FITS_CONVENTION)
        # TODO(MatthewJA): Remove (or disable) this sanity check.

        # The bbox is backwards along the x-axis for some reason.
        bbox[0] = bbox[0][::-1]
        assert bbox[0][0] < bbox[0][1]
        assert bbox[1][0] < bbox[1][1]

        bbox = numpy.array(bbox)

        # What is this radio source called? Check if we have an object in the
        # bounding box. We'll cache these results because there is a lot of
        # overlap.
        cache_key = tuple(tuple(b) for b in bbox)
        if cache_key in bbox_cache_:
            index = bbox_cache_[cache_key]
        else:
            x_gt_min = atlas_positions[:, 0] >= bbox[0, 0]
            x_lt_max = atlas_positions[:, 0] <= bbox[0, 1]
            y_gt_min = atlas_positions[:, 1] >= bbox[1, 0]
            y_lt_max = atlas_positions[:, 1] <= bbox[1, 1]
            within = numpy.all([x_gt_min, x_lt_max, y_gt_min, y_lt_max], axis=0)
            indices = numpy.where(within)[0]

            if len(indices) == 0:
                logging.debug('Skipping radio source not in catalogue for '
                              '%s', zooniverse_id)
                continue
            else:
                if len(indices) > 1:
                    logging.debug('Found multiple (%d) ATLAS matches '
                                  'for %s', len(indices), zooniverse_id)

                index = indices[0]

            bbox_cache_[cache_key] = index

        atlas_ids.append(str(index))

    atlas_ids.sort()

    if not atlas_ids:
        raise CatalogueError('No catalogued radio sources.')

    return ';'.join(atlas_ids)


def parse_classification(classification, subject, atlas_positions, wcs,
                         pix_offset):
    """Converts a raw RGZ classification into a classification dict.

    Scales all positions and flips y axis of clicks.

    classification: RGZ classification dict.
    subject: Associated RGZ subject dict.
    atlas_positions: [[RA, DEC]] NumPy array.
    wcs: World coordinate system of the ATLAS image.
    pix_offset: (x, y) pixel position of this radio subject on the ATLAS image.
    -> dict mapping radio signature to corresponding IR host pixel location
    """
    result = {}

    n_invalid = 0

    for annotation in classification['annotations']:
        if 'radio' not in annotation:
            # This is a metadata annotation and we can ignore it.
            continue

        if annotation['radio'] == 'No Contours':
            # I'm not sure how this occurs. I'm going to ignore it.
            continue

        try:
            radio_signature = make_radio_combination_signature(
                    annotation['radio'], wcs, atlas_positions,
                    subject, pix_offset)
        except CatalogueError:
            # Ignore invalid annotations.
            n_invalid += 1
            logging.debug('Ignoring invalid annotation for %s.',
                          subject['zooniverse_id'])
            continue

        if annotation['ir'] == 'No Sources':
            ir_location = (None, None)
        else:
            ir_x = float(annotation['ir']['0']['x'])
            ir_y = float(annotation['ir']['0']['y'])

            # Rescale to a consistent size.
            ir_x *= config['surveys']['atlas']['click_to_fits_x']
            ir_y *= config['surveys']['atlas']['click_to_fits_y']

            # Ignore out-of-range data.
            if not 0 <= ir_x <= config['surveys']['atlas']['fits_width']:
                n_invalid += 1
                continue

            if not 0 <= ir_y <= config['surveys']['atlas']['fits_height']:
                n_invalid += 1
                continue

            # Flip the y axis to match other data conventions.
            ir_y = config['surveys']['atlas']['fits_height'] - ir_y

            # Rescale to match the mosaic WCS.
            ir_x *= config['surveys']['atlas']['mosaic_scale_x']
            ir_y *= config['surveys']['atlas']['mosaic_scale_y']

            # Move to the reference location of the radio subject.
            ir_x += pix_offset[0]
            ir_y += pix_offset[1]

            # Convert the location into RA/DEC.
            (ir_x,), (ir_y,) = wcs.wcs_pix2world([ir_x], [ir_y], 1)

            ir_location = (ir_x, ir_y)

        result[radio_signature] = ir_location

    if n_invalid:
        logging.debug('%d invalid annotations for %s.', n_invalid,
                      subject['zooniverse_id'])

    return result


def import_classifications(f_h5, test=False):
    """Imports Radio Galaxy Zoo classifications into crowdastro.

    f_h5: An HDF5 file.
    test: Flag to run on only 10 subjects. Default False.
    """
    # TODO(MatthewJA): This only works for ATLAS/CDFS. Generalise.
    from . import rgz_data as data
    atlas_positions = f_h5['/atlas/cdfs/numeric'][:, :2]
    atlas_ids = f_h5['/atlas/cdfs/string']['zooniverse_id']
    classification_positions = []
    classification_combinations = []
    classification_usernames = []

    with astropy.io.fits.open(config['data_sources']['atlas_image'],
                              ignore_blank=True) as atlas_image:
        wcs = astropy.wcs.WCS(atlas_image[0].header).dropaxis(3).dropaxis(2)

    for obj_index, atlas_id in enumerate(atlas_ids):
        subject = data.get_subject(atlas_id.decode('ascii'))
        assert subject['zooniverse_id'] == atlas_ids[obj_index].decode('ascii')
        classifications = data.get_subject_classifications(subject)
        offset, = wcs.all_world2pix([subject['coords']], FITS_CONVENTION)
        # The coords are of the middle of the subject.
        offset[0] -= (config['surveys']['atlas']['fits_width'] *
                      config['surveys']['atlas']['mosaic_scale_x'] // 2)
        offset[1] -= (config['surveys']['atlas']['fits_height'] *
                      config['surveys']['atlas']['mosaic_scale_y'] // 2)

        for c_index, classification in enumerate(classifications):
            user_name = classification.get('user_name', '').encode(
                    'ascii', errors='ignore')
            # Usernames actually don't have an upper length limit on RGZ(?!) so
            # I'll cap everything at 50 characters for my own sanity.
            if len(user_name) > 50:
                user_name = user_name[:50]

            classification = parse_classification(classification, subject,
                                                  atlas_positions, wcs, offset)
            full_radio = '|'.join(classification.keys())
            for radio, location in classification.items():
                pos_row = (obj_index, location[0], location[1])
                com_row = (obj_index, full_radio, radio)
                # A little redundancy here with the index, but we can assert
                # that they are the same later to check integrity.
                classification_positions.append(pos_row)
                classification_combinations.append(com_row)
                classification_usernames.append(user_name)

    combinations_dtype = [('index', 'int'),
                          ('full_signature', '<S{}'.format(
                                    MAX_RADIO_SIGNATURE_LENGTH)),
                          ('signature', '<S{}'.format(
                                    MAX_RADIO_SIGNATURE_LENGTH))]
    classification_positions = numpy.array(classification_positions,
                                           dtype=float)
    classification_combinations = numpy.array(classification_combinations,
                                              dtype=combinations_dtype)

    f_h5['/atlas/cdfs/'].create_dataset('classification_positions',
                                        data=classification_positions,
                                        dtype=float)
    f_h5['/atlas/cdfs/'].create_dataset('classification_usernames',
                                        data=classification_usernames,
                                        dtype='<S50')
    f_h5['/atlas/cdfs/'].create_dataset('classification_combinations',
                                        data=classification_combinations,
                                        dtype=combinations_dtype)


def _populate_parser(parser):
    parser.description = 'Imports and standardises data into crowdastro.'
    parser.add_argument('--h5', default='data/crowdastro.h5',
                        help='HDF5 output file')
    parser.add_argument('--test', action='store_true', default=False,
                        help='Run with a small number of subjects',)
    parser.add_argument('--ir', choices={'swire', 'wise'},
                        default='swire', help='which infrared survey to use')


def _main(args):
    with h5py.File(args.h5, 'w') as f_h5:
        prep_h5(f_h5, args.ir)
        import_atlas(f_h5, test=args.test)
        if args.ir == 'swire':
            import_swire(f_h5)
        elif args.ir == 'wise':
            import_wise(f_h5)
        import_norris(f_h5)
        import_fan(f_h5)
        import_classifications(f_h5)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Generates training data (potential hosts and their astronomical features).

Also generates training/testing indices.

Matthew Alger
The Australian National University
2016
"""

import argparse

import h5py
import numpy
import sklearn.neighbors

from .config import config

FITS_HEIGHT = config['surveys']['atlas']['fits_height']
FITS_WIDTH = config['surveys']['atlas']['fits_height']
IMAGE_SIZE = FITS_HEIGHT * FITS_WIDTH  # px
ARCMIN = 1 / 60  # deg


def remove_nans(n):
    """Replaces NaN with 0."""
    if numpy.ma.is_masked(n):
        return 0

    return float(n)


def generate(f_h5, out_f_h5):
    """Generates potential hosts and their astronomical features.

    f_h5: crowdastro input HDF5 file.
    out_f_h5: Training data output HDF5 file.
    """
    if f_h5.attrs['ir_survey'] == 'swire':
        swire = f_h5['/swire/cdfs/numeric']
        astro_features = swire[:, 2:7]
        distances = swire[:, 7].reshape((-1, 1))
        images = swire[:, 8:]
        coords = swire[:, :2]
    elif f_h5.attrs['ir_survey'] == 'wise':
        wise = f_h5['/wise/cdfs/numeric']
        magnitudes = wise[:, 2:6]
        distances = wise[:, 6].reshape((-1, 1))
        images = wise[:, 7:]
        coords = wise[:, :2]

        # Magnitude differences are probably useful features.
        w1_w2 = magnitudes[:, 0] - magnitudes[:, 1]
        w2_w3 = magnitudes[:, 1] - magnitudes[:, 2]

        # Converting the magnitudes to a linear scale seems to improve
        # performance.
        linearised_magnitudes = numpy.power(10, -0.4 * magnitudes)
        w1_w2 = numpy.power(10, -0.4 * w1_w2)
        w2_w3 = numpy.power(10, -0.4 * w2_w3)
        astro_features = numpy.concatenate([linearised_magnitudes,
                w1_w2.reshape((-1, 1)),
                w2_w3.reshape((-1, 1))], axis=1)


    # We now need to find the labels for each.
    truths = set(f_h5['/atlas/cdfs/consensus_objects'][:, 1])
    labels = numpy.array([o in truths for o in range(len(astro_features))])

    assert len(labels) == len(astro_features)
    assert len(astro_features) == len(distances)
    assert len(distances) == len(images)

    features = numpy.hstack([astro_features, distances, images])
    n_astro = features.shape[1] - images.shape[1]

    # Save to HDF5.
    out_f_h5.create_dataset('labels', data=labels)
    out_f_h5.create_dataset('features', data=features)
    out_f_h5.create_dataset('positions', data=coords)
    out_f_h5.attrs['ir_survey'] = f_h5.attrs['ir_survey']

    # These testing sets are used for training the CNN. For training/testing
    # models, use crowdastro.generate_test_sets.

    # We want to ensure our training set is never in our testing set, so
    # 1. assign all ATLAS objects to a train or test set,
    # 2. if a IR object is nearby a testing ATLAS object, assign it to a test
    #    set, and
    # 3. assign all other IR objects to a train set.
    n_atlas = f_h5['/atlas/cdfs/numeric'].shape[0]
    indices = numpy.arange(n_atlas)
    numpy.random.shuffle(indices)
    atlas_test_indices = indices[:int(n_atlas * config['test_size'])]
    atlas_train_indices = indices[int(n_atlas * config['test_size']):]

    atlas_test_indices.sort()
    atlas_train_indices.sort()

    is_atlas_train = numpy.zeros((n_atlas,))
    is_atlas_test = numpy.zeros((n_atlas,))

    is_atlas_test[atlas_test_indices] = 1
    is_atlas_train[atlas_train_indices] = 1

    n_ir = len(astro_features)
    is_ir_train = numpy.ones((n_ir))
    is_ir_test = numpy.zeros((n_ir))

    for atlas_index in atlas_test_indices:
        ir = f_h5['/atlas/cdfs/numeric'][atlas_index, n_astro + IMAGE_SIZE:]
        nearby = (ir < ARCMIN).nonzero()[0]
        for ir_index in nearby:
            is_ir_test[ir_index] = 1
            is_ir_train[ir_index] = 0

    out_f_h5.create_dataset('is_atlas_train', data=is_atlas_train.astype(bool))
    out_f_h5.create_dataset('is_atlas_test', data=is_atlas_test.astype(bool))
    out_f_h5.create_dataset('is_ir_train', data=is_ir_train.astype(bool))
    out_f_h5.create_dataset('is_ir_test', data=is_ir_test.astype(bool))


def _populate_parser(parser):
    parser.description = 'Generates training data (potential hosts and their ' \
                         'astronomical features).'
    parser.add_argument('-i', default='data/crowdastro.h5',
                        help='HDF5 input file')
    parser.add_argument('-o', default='data/training.h5',
                        help='HDF5 output file')


def _main(args):
    with h5py.File(args.i, 'r') as f_h5:
        assert f_h5.attrs['version'] == '0.5.1'
        with h5py.File(args.o, 'w') as out_f_h5:
            generate(f_h5, out_f_h5)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



from astropy.io import ascii
import numpy as n
import math as m
import scipy
import csv

#----------------------------------------------------------------------------------
def ra2deg(hh,mm,ss):
  # converts RA from hh:mm:ss to decimal degrees

  # check values
  if hh > 24:
    print 'ERROR: RA must not exceed 24 hours\n'
    print 'exiting'

  radeg = (((float(ss)/60.) + float(mm))/60.) + float(hh)
  radeg = 15. * radeg

  return radeg

def dec2deg(dd,mm,ss):
  # converts DEC from dd:mm:ss to decimal degrees
  dd = float(dd)
  sign = "+"
  # check values
  if dd > 90:
    print 'ERROR: DEC must not exceed 90 degrees\n'
    print 'exiting'

  if dd < -90:
    print 'ERROR: DEC must not exceed -90 degrees\n'
    print 'exiting'

  # check sign of declination

  if dd < 0:
    sign = "-"
  
  decdeg = (((float(ss)/60.) + float(mm))/60.) + abs(dd)

  if sign == '-':
    decdeg = decdeg * -1

  return decdeg


#----------------------------------------------------------------------------------
# create output files
outhost = open('RGZ-ATLAS-CDFS-hosts.csv','w')

writer = csv.writer(outhost)
writer.writerow( ('zooniverse_id', 'rg_id','RA_H','RA_M','RA_S','DEC_D','DEC_M','DEC_S',\
  'RA_DEG','DEC_DEG','Agreement','S2','S2_err','Ncomp','SWIRE_ID','ra','dec','unc_ra','unc_dec',\
  'tile','flux_ap1_36','uncf_ap1_36','flux_ap2_36','uncf_ap2_36','flux_ap3_36','uncf_ap3_36',\
  'flux_ap4_36','uncf_ap4_36','flux_ap5_36','uncf_ap5_36','flux_kr_36','uncf_kr_36',\
  'rad_kr_36','flux_iso_36','uncf_iso_36','area_iso_36','stell_36','a_36','b_36','theta_36',\
  'flgs_36','ext_fl_36','cov_avg_36','cov_36','flux_ap1_45','uncf_ap1_45','flux_ap2_45',\
  'uncf_ap2_45','flux_ap3_45','uncf_ap3_45','flux_ap4_45','uncf_ap4_45','flux_ap5_45','uncf_ap5_45',\
  'flux_kr_45','uncf_kr_45','rad_kr_45','flux_iso_45','uncf_iso_45','area_iso_45','stell_45','a_45',\
  'b_45','theta_45','flgs_45','ext_fl_45','cov_avg_45','cov_45','flux_ap1_58','uncf_ap1_58',\
  'flux_ap2_58','uncf_ap2_58','flux_ap3_58','uncf_ap3_58','flux_ap4_58','uncf_ap4_58','flux_ap5_58',\
  'uncf_ap5_58','flux_kr_58',' uncf_kr_58','rad_kr_58','flux_iso_58','uncf_iso_58','area_iso_58',\
  'stell_58','a_58','b_58','theta_58','flgs_58','ext_fl_58','cov_avg_58','cov_58','flux_ap1_80',\
  'uncf_ap1_80','flux_ap2_80','uncf_ap2_80','flux_ap3_80','uncf_ap3_80','flux_ap4_80','uncf_ap4_80',\
  'flux_ap5_80','uncf_ap5_80','flux_kr_80','uncf_kr_80','rad_kr_80','flux_iso_80','uncf_iso_80',\
  'area_iso_80','stell_80','a_80','b_80','theta_80','flgs_80','ext_fl_80','cov_avg_80','cov_80',\
  'detid_24','flux_ap2_24','uncf_ap2_24','flux_ap3_24','uncf_ap3_24','flux_ap4_24','uncf_ap4_24',\
  'flux_ap5_24','uncf_ap5_24','flux_kr_24','uncf_kr_24','rad_kr_24','flux_iso_24','uncf_iso_24',\
  'area_iso_24','stell_24','a_24','b_24','theta_24','flgs_24','ext_fl_24','cov_avg_24','cov_24',\
  'fl_2mass','cov_u','cov_g','cov_r','cov_i','cov_z','ra_opt','dec_opt','ap_m_u','msig_u',\
  'ap_m_g','msig_g','ap_m_r','msig_r','ap_m_i','msig_i','ap_m_z','msig_z','int_m_u','int_m_g',\
  'int_m_r','int_m_i','int_m_z','fl_u','fl_g','fl_r','fl_i','fl_z','xid_p','xid_dist'))


outcomp = open('RGZ-ATLAS-CDFS-components.csv','w')
writer2 = csv.writer(outcomp)
writer2.writerow(('rgz_id','RA_H','RA_M','RA_S','DEC_D','DEC_M','DEC_S',\
  'RA_DEG','DEC_DEG','ATLAS_ID','ATLAS_Name','ATLAS_RA','ATLAS_DEC','ATLAS_RA_deg','ATLAS_DEC_deg',\
  'RA_err','DEC_err','rms','BWS','Obs_freq','Sp','Sp_err','S','S_err','Sp2','Sp2_err','S2','S2_err',\
  'Deconv','Deconv_err','class','Sindex','Sindex_err','field','Agreement'))

# SWIRE catalogue
# -- for some reason astropy can't read in the SWIRE ipac table
# -- added # to the ipac comments and column headers
swirefile = 'SWIRE3_CDFS_cat_IRAC24_21Dec05.tbl'
swiredata = ascii.read(swirefile)#,format='ipac',definition='ignore',guess=False)

swireid = n.array(swiredata['col2'])

# atlas RGZ files
hostfile = 'rgz_hosts_kde_24MAY2016.csv'
compfile = 'rgz_radio_components_kde_24MAY2106.csv'

# ATLAS catalogue from Franzen+(2015)
atlasfile = 'ATLASDR3_cmpcat_23July2015.csv'
atlasdata = ascii.read(atlasfile)

atlasid = atlasdata['ID']
atlasra = n.array(atlasdata['RA_deg'])
atlasdec = n.array(atlasdata['Dec_deg'])

# read in files
hostdata = ascii.read(hostfile)

zooid = hostdata['zooniverse_id']
source = hostdata['source']
rgzname = hostdata['rgz_name']
swirename = hostdata['swire_name']
ra = hostdata['ra']
dec = hostdata['dec']
agreement = hostdata['agreement']

compdata = ascii.read(compfile)
comprgzname = compdata['rgz_name']
compid = compdata['radio_component']
compagree = compdata['agreement']


# split RA and DEC values
rah = n.zeros(len(ra))
ram = n.zeros(len(ra))
ras = n.zeros(len(ra))
decd = n.zeros(len(dec))
decm = n.zeros(len(dec))
decs = n.zeros(len(dec))

ii = 0
while ii < len(ra):
  ## RA
  tmp = ra[ii].split("h")
  rah[ii] = int(tmp[0])

  tmp2 = tmp[1].split("m")
  ram[ii] = int(tmp2[0])

  tmp3 = tmp2[1].split("s")
  ras[ii] = float(tmp3[0])

  # convert to decimal degrees
  raindeg = ra2deg(rah[ii],ram[ii],ras[ii])

  tmp = ''
  tmp2 = ''
  tmp3 = ''
  
  ## DEC
  tmp = dec[ii].split("d")
  decd[ii] = int(tmp[0])

  tmp2 = tmp[1].split("m")
  decm[ii] = int(tmp2[0])

  tmp3 = tmp2[1].split("s")
  decs[ii] = float(tmp3[0])

  # convert to decimal degrees
  decindeg = dec2deg(decd[ii],decm[ii],decs[ii])
  tmp = ''
  tmp2 = ''
  tmp3 = ''

  # determine how many components there are for this RGZ subject
  indcomp = n.where(rgzname[ii] == comprgzname)
  ncomp = len(indcomp[0])

  # only one radio emitting structure for the host galaxy
  if ncomp == 1:
    # write the values out to the host catalogue --> this includes SWIRE values
    newind = n.where(compid[indcomp][0] == atlasdata['ID'])

    # locate SWIRE id in SWIRE catalogue
    swind = n.where(swirename[ii] == swireid)
    if (len(swind[0]) < 1) or (len(swind[0]) > 2):
      print 'SWIRE id invalid...exiting...'
      print ''
      exit()

    writer.writerow((zooid[ii], rgzname[ii], rah[ii],\
      ram[ii], ras[ii], decd[ii], decm[ii], decs[ii], raindeg, decindeg, agreement[ii], \
      atlasdata['S2'][newind][0], atlasdata['S2_err'][newind][0], ncomp, \
      swiredata['col2'][swind][0], swiredata['col3'][swind][0], \
      swiredata['col4'][swind][0], swiredata['col5'][swind][0], swiredata['col6'][swind][0], \
      swiredata['col7'][swind][0],  swiredata['col8'][swind][0], swiredata['col9'][swind][0],\
      swiredata['col10'][swind][0], swiredata['col11'][swind][0], swiredata['col12'][swind][0], \
      swiredata['col13'][swind][0], swiredata['col14'][swind][0], swiredata['col15'][swind][0], \
      swiredata['col16'][swind][0], swiredata['col17'][swind][0], swiredata['col18'][swind][0], \
      swiredata['col19'][swind][0], swiredata['col20'][swind][0], swiredata['col21'][swind][0], \
      swiredata['col22'][swind][0], swiredata['col23'][swind][0], swiredata['col24'][swind][0], \
      swiredata['col25'][swind][0], swiredata['col26'][swind][0], swiredata['col27'][swind][0], \
      swiredata['col28'][swind][0], swiredata['col29'][swind][0], swiredata['col30'][swind][0], \
      swiredata['col31'][swind][0], swiredata['col32'][swind][0], swiredata['col33'][swind][0], \
      swiredata['col34'][swind][0], swiredata['col35'][swind][0], swiredata['col36'][swind][0], \
      swiredata['col37'][swind][0], swiredata['col38'][swind][0], swiredata['col39'][swind][0], \
      swiredata['col40'][swind][0], swiredata['col41'][swind][0], swiredata['col42'][swind][0], \
      swiredata['col43'][swind][0], swiredata['col44'][swind][0], swiredata['col45'][swind][0], \
      swiredata['col46'][swind][0], swiredata['col47'][swind][0], swiredata['col48'][swind][0], \
      swiredata['col49'][swind][0], swiredata['col50'][swind][0], swiredata['col51'][swind][0], \
      swiredata['col52'][swind][0], swiredata['col53'][swind][0], swiredata['col54'][swind][0], \
      swiredata['col55'][swind][0], swiredata['col56'][swind][0], swiredata['col57'][swind][0], \
      swiredata['col58'][swind][0], swiredata['col59'][swind][0], swiredata['col60'][swind][0], \
      swiredata['col61'][swind][0], swiredata['col62'][swind][0], swiredata['col63'][swind][0], \
      swiredata['col64'][swind][0], swiredata['col65'][swind][0], swiredata['col66'][swind][0], \
      swiredata['col67'][swind][0], swiredata['col68'][swind][0], swiredata['col69'][swind][0], \
      swiredata['col70'][swind][0], swiredata['col71'][swind][0], swiredata['col72'][swind][0], \
      swiredata['col73'][swind][0], swiredata['col74'][swind][0], swiredata['col75'][swind][0], \
      swiredata['col76'][swind][0], swiredata['col77'][swind][0], swiredata['col78'][swind][0], \
      swiredata['col79'][swind][0], swiredata['col80'][swind][0], swiredata['col81'][swind][0], \
      swiredata['col82'][swind][0], swiredata['col83'][swind][0], swiredata['col84'][swind][0], \
      swiredata['col85'][swind][0], swiredata['col86'][swind][0], swiredata['col87'][swind][0], \
      swiredata['col88'][swind][0], swiredata['col89'][swind][0], swiredata['col90'][swind][0], \
      swiredata['col91'][swind][0], swiredata['col92'][swind][0], swiredata['col93'][swind][0], \
      swiredata['col94'][swind][0], swiredata['col95'][swind][0], swiredata['col96'][swind][0], \
      swiredata['col97'][swind][0], swiredata['col98'][swind][0], swiredata['col99'][swind][0], \
      swiredata['col100'][swind][0], swiredata['col101'][swind][0], swiredata['col102'][swind][0], \
      swiredata['col103'][swind][0], swiredata['col104'][swind][0], swiredata['col105'][swind][0], \
      swiredata['col106'][swind][0], swiredata['col107'][swind][0], swiredata['col108'][swind][0], \
      swiredata['col109'][swind][0], swiredata['col110'][swind][0], swiredata['col111'][swind][0], \
      swiredata['col112'][swind][0], swiredata['col113'][swind][0], swiredata['col114'][swind][0], \
      swiredata['col115'][swind][0], swiredata['col116'][swind][0], swiredata['col117'][swind][0], \
      swiredata['col118'][swind][0], swiredata['col119'][swind][0], swiredata['col120'][swind][0], \
      swiredata['col121'][swind][0], swiredata['col122'][swind][0], swiredata['col123'][swind][0], \
      swiredata['col124'][swind][0], swiredata['col125'][swind][0], swiredata['col126'][swind][0], \
      swiredata['col127'][swind][0], swiredata['col128'][swind][0], swiredata['col129'][swind][0], \
      swiredata['col130'][swind][0], swiredata['col131'][swind][0], swiredata['col132'][swind][0], \
      swiredata['col133'][swind][0], swiredata['col134'][swind][0], swiredata['col135'][swind][0], \
      swiredata['col136'][swind][0], swiredata['col137'][swind][0], swiredata['col138'][swind][0], \
      swiredata['col139'][swind][0], swiredata['col140'][swind][0], swiredata['col141'][swind][0], \
      swiredata['col142'][swind][0], swiredata['col143'][swind][0], swiredata['col144'][swind][0], \
      swiredata['col145'][swind][0], swiredata['col146'][swind][0], swiredata['col147'][swind][0], \
      swiredata['col148'][swind][0], swiredata['col149'][swind][0], swiredata['col150'][swind][0], \
      swiredata['col151'][swind][0], swiredata['col152'][swind][0], swiredata['col153'][swind][0], \
      swiredata['col154'][swind][0], swiredata['col155'][swind][0], swiredata['col156'][swind][0]))


    # write the values out to the component catalogue --> only ATLAS data
    writer2.writerow((rgzname[ii],rah[ii],ram[ii],ras[ii],decd[ii],decm[ii],decs[ii],raindeg,decindeg,atlasdata['ID'][newind][0],atlasdata['name'][newind][0],atlasdata['RA'][newind][0],atlasdata['Dec'][newind][0],atlasdata['RA_deg'][newind][0],atlasdata['Dec_deg'][newind][0],atlasdata['RA_err'][newind][0],atlasdata['Dec_err'][newind][0],atlasdata['rms'][newind][0],atlasdata['BWS'][newind][0],atlasdata['Obs_freq'][newind][0],atlasdata['Sp'][newind][0],atlasdata['Sp_err'][newind][0],atlasdata['S'][newind][0],atlasdata['S_err'][newind][0],atlasdata['Sp2'][newind][0],atlasdata['Sp2_err'][newind][0],atlasdata['S2'][newind][0],atlasdata['S2_err'][newind][0],atlasdata['Deconv'][newind][0],atlasdata['Deconv_err'][newind][0],atlasdata['class'][newind][0],atlasdata['Sindex'][newind][0],atlasdata['Sindex_err'][newind][0],atlasdata['field'][newind][0],compagree[ii]))


  # more then one radio emitting structure for the host galaxy
  if ncomp > 1:
    newflux = 0.0
    newfluxerr = 0.0
    for jj in n.arange(ncomp):
      newind = indcomp[0][jj]
      atlasind = n.where(compid[newind] == atlasdata['ID'])
    
      # determine the total flux for the radio emitting source
      newflux = atlasdata['S2'][atlasind][0] + newflux
      newfluxerr = (atlasdata['S2_err'][atlasind][0]*atlasdata['S2_err'][atlasind][0]) + newfluxerr

      ### need to calculate the error in the flux ###

      # write each radio component to the component catalogue --> only ATLAS data
      writer2.writerow((rgzname[ii],rah[ii],ram[ii],ras[ii],decd[ii],decm[ii],decs[ii],raindeg,decindeg,atlasdata['ID'][atlasind][0],atlasdata['name'][atlasind][0],atlasdata['RA'][atlasind][0],atlasdata['Dec'][atlasind][0],atlasdata['RA_deg'][atlasind][0],atlasdata['Dec_deg'][atlasind][0],atlasdata['RA_err'][atlasind][0],atlasdata['Dec_err'][atlasind][0],atlasdata['rms'][atlasind][0],atlasdata['BWS'][atlasind][0],atlasdata['Obs_freq'][atlasind][0],atlasdata['Sp'][atlasind][0],atlasdata['Sp_err'][atlasind][0],atlasdata['S'][atlasind][0],atlasdata['S_err'][atlasind][0],atlasdata['Sp2'][atlasind][0],atlasdata['Sp2_err'][atlasind][0],atlasdata['S2'][atlasind][0],atlasdata['S2_err'][atlasind][0],atlasdata['Deconv'][atlasind][0],atlasdata['Deconv_err'][atlasind][0],atlasdata['class'][atlasind][0],atlasdata['Sindex'][atlasind][0],atlasdata['Sindex_err'][atlasind][0],atlasdata['field'][atlasind][0],compagree[newind]))


    # write the combined radio components to the host file

    # locate SWIRE id in SWIRE catalogue
    swind = n.where(swirename[ii] == swireid)
    if (len(swind[0]) < 1) or (len(swind[0]) > 2):
      print 'SWIRE id invalid...exiting...'
      print ''
      exit()

    newfluxerr = n.sqrt(newfluxerr)

    writer.writerow((zooid[ii], rgzname[ii], rah[ii],\
      ram[ii], ras[ii], decd[ii], decm[ii], decs[ii], raindeg, decindeg, agreement[ii], \
      newflux, newfluxerr, ncomp, swiredata['col2'][swind][0], swiredata['col3'][swind][0], \
      swiredata['col4'][swind][0], swiredata['col5'][swind][0], swiredata['col6'][swind][0], \
      swiredata['col7'][swind][0],  swiredata['col8'][swind][0], swiredata['col9'][swind][0],\
      swiredata['col10'][swind][0], swiredata['col11'][swind][0], swiredata['col12'][swind][0], \
      swiredata['col13'][swind][0], swiredata['col14'][swind][0], swiredata['col15'][swind][0], \
      swiredata['col16'][swind][0], swiredata['col17'][swind][0], swiredata['col18'][swind][0], \
      swiredata['col19'][swind][0], swiredata['col20'][swind][0], swiredata['col21'][swind][0], \
      swiredata['col22'][swind][0], swiredata['col23'][swind][0], swiredata['col24'][swind][0], \
      swiredata['col25'][swind][0], swiredata['col26'][swind][0], swiredata['col27'][swind][0], \
      swiredata['col28'][swind][0], swiredata['col29'][swind][0], swiredata['col30'][swind][0], \
      swiredata['col31'][swind][0], swiredata['col32'][swind][0], swiredata['col33'][swind][0], \
      swiredata['col34'][swind][0], swiredata['col35'][swind][0], swiredata['col36'][swind][0], \
      swiredata['col37'][swind][0], swiredata['col38'][swind][0], swiredata['col39'][swind][0], \
      swiredata['col40'][swind][0], swiredata['col41'][swind][0], swiredata['col42'][swind][0], \
      swiredata['col43'][swind][0], swiredata['col44'][swind][0], swiredata['col45'][swind][0], \
      swiredata['col46'][swind][0], swiredata['col47'][swind][0], swiredata['col48'][swind][0], \
      swiredata['col49'][swind][0], swiredata['col50'][swind][0], swiredata['col51'][swind][0], \
      swiredata['col52'][swind][0], swiredata['col53'][swind][0], swiredata['col54'][swind][0], \
      swiredata['col55'][swind][0], swiredata['col56'][swind][0], swiredata['col57'][swind][0], \
      swiredata['col58'][swind][0], swiredata['col59'][swind][0], swiredata['col60'][swind][0], \
      swiredata['col61'][swind][0], swiredata['col62'][swind][0], swiredata['col63'][swind][0], \
      swiredata['col64'][swind][0], swiredata['col65'][swind][0], swiredata['col66'][swind][0], \
      swiredata['col67'][swind][0], swiredata['col68'][swind][0], swiredata['col69'][swind][0], \
      swiredata['col70'][swind][0], swiredata['col71'][swind][0], swiredata['col72'][swind][0], \
      swiredata['col73'][swind][0], swiredata['col74'][swind][0], swiredata['col75'][swind][0], \
      swiredata['col76'][swind][0], swiredata['col77'][swind][0], swiredata['col78'][swind][0], \
      swiredata['col79'][swind][0], swiredata['col80'][swind][0], swiredata['col81'][swind][0], \
      swiredata['col82'][swind][0], swiredata['col83'][swind][0], swiredata['col84'][swind][0], \
      swiredata['col85'][swind][0], swiredata['col86'][swind][0], swiredata['col87'][swind][0], \
      swiredata['col88'][swind][0], swiredata['col89'][swind][0], swiredata['col90'][swind][0], \
      swiredata['col91'][swind][0], swiredata['col92'][swind][0], swiredata['col93'][swind][0], \
      swiredata['col94'][swind][0], swiredata['col95'][swind][0], swiredata['col96'][swind][0], \
      swiredata['col97'][swind][0], swiredata['col98'][swind][0], swiredata['col99'][swind][0], \
      swiredata['col100'][swind][0], swiredata['col101'][swind][0], swiredata['col102'][swind][0], \
      swiredata['col103'][swind][0], swiredata['col104'][swind][0], swiredata['col105'][swind][0], \
      swiredata['col106'][swind][0], swiredata['col107'][swind][0], swiredata['col108'][swind][0], \
      swiredata['col109'][swind][0], swiredata['col110'][swind][0], swiredata['col111'][swind][0], \
      swiredata['col112'][swind][0], swiredata['col113'][swind][0], swiredata['col114'][swind][0], \
      swiredata['col115'][swind][0], swiredata['col116'][swind][0], swiredata['col117'][swind][0], \
      swiredata['col118'][swind][0], swiredata['col119'][swind][0], swiredata['col120'][swind][0], \
      swiredata['col121'][swind][0], swiredata['col122'][swind][0], swiredata['col123'][swind][0], \
      swiredata['col124'][swind][0], swiredata['col125'][swind][0], swiredata['col126'][swind][0], \
      swiredata['col127'][swind][0], swiredata['col128'][swind][0], swiredata['col129'][swind][0], \
      swiredata['col130'][swind][0], swiredata['col131'][swind][0], swiredata['col132'][swind][0], \
      swiredata['col133'][swind][0], swiredata['col134'][swind][0], swiredata['col135'][swind][0], \
      swiredata['col136'][swind][0], swiredata['col137'][swind][0], swiredata['col138'][swind][0], \
      swiredata['col139'][swind][0], swiredata['col140'][swind][0], swiredata['col141'][swind][0], \
      swiredata['col142'][swind][0], swiredata['col143'][swind][0], swiredata['col144'][swind][0], \
      swiredata['col145'][swind][0], swiredata['col146'][swind][0], swiredata['col147'][swind][0], \
      swiredata['col148'][swind][0], swiredata['col149'][swind][0], swiredata['col150'][swind][0], \
      swiredata['col151'][swind][0], swiredata['col152'][swind][0], swiredata['col153'][swind][0], \
      swiredata['col154'][swind][0], swiredata['col155'][swind][0], swiredata['col156'][swind][0]))


  # no radio components???
  if ncomp < 1:
    print 'no components for this RGZ subject????'
    print ''
    print rgzname[ii],agreement[ii],ncomp
    print ''
    print 'exiting'
    exit()

 
  ii = ii + 1


outhost.close()
outcomp.close()




# -*- coding: utf-8 -*-

"""Utilities for interacting with the Radio Galaxy Zoo database.

Matthew Alger
The Australian National University
2016
"""

import io
import os.path

import astropy.io.fits
import astropy.io.votable
import astropy.wcs
import matplotlib.colors
import matplotlib.pyplot
import numpy
import pymongo
import requests

from .config import config

client = pymongo.MongoClient(config['mongo']['host'], config['mongo']['port'])
db = client[config['data_sources']['radio_galaxy_zoo_db']]


def require_atlas(f):
    """Decorator that ensures a subject (the first argument) is from the ATLAS
    survey.
    """
    def g(subject, *args, **kwargs):
        if subject['metadata']['survey'] != 'atlas':
            raise ValueError('Subject not from ATLAS survey.')

        return f(subject, *args, **kwargs)

    return g


def get_random_subject():
    return list(db.radio_subjects.aggregate([
        {'$match': {'metadata.survey': 'atlas'}},
        {'$sample': {'size': 1}}]))[0]


def get_subject(zid):
    """Gets a Radio Galaxy Zoo subject from the database.

    zid: Zooniverse ID of subject.
    -> RGZ subject dict.
    """
    return db.radio_subjects.find_one({'zooniverse_id': zid})


def get_subject_raw_id(subject_id):
    """Gets a Radio Galaxy Zoo subject from the database using its subject ID.

    subject_id: Raw ID of subject.
    -> RGZ subject dict.
    """
    return db.radio_subjects.find_one({'_id': subject_id})


def get_all_classifications():
    """Returns cursor yielding all RGZ classification dicts.

    -> MongoDB cursor
    """
    return db.radio_classifications.find()


def get_subject_classifications(subject):
    """Returns cursor yielding all classifications associated with a subject.

    subject: RGZ subject dict.
    -> MongoDB cursor
    """
    return db.radio_classifications.find({'subject_ids': subject['_id']})


def get_all_subjects(survey=None, field=None):
    """Returns cursor yielding RGZ subject dicts.

    survey: Optional. Survey subject was observed in. {'atlas', None}.
    field: Optional. Field subject was observed in. {'cdfs', 'elais-s1', None}.
    -> MongoDB cursor
    """
    if survey is None:
        return db.radio_subjects.find()

    if survey == 'atlas':
        if field is None:
            return db.radio_subjects.find(
                    {'metadata.survey': 'atlas'})

        if field == 'cdfs':
            return db.radio_subjects.find(
                    {'metadata.survey': 'atlas',
                     'metadata.source': {'$regex': '^CI'}})

        if field == 'elais-s1':
            return db.radio_subjects.find(
                    {'metadata.survey': 'atlas',
                     'metadata.source': {'$regex': '^EI'}})

        raise ValueError('Unknown field: {}'.format(field))

    raise ValueError('Unknown survey: {}'.format(survey))


@require_atlas
def open_fits(subject, field, wavelength, size='2x2'):
    """Opens a FITS image of a subject.

    Can be used as a context handler.

    subject: RGZ subject dict, from the ATLAS survey.
    field: 'elais' or 'cdfs'
    wavelength: 'ir' or 'radio'
    size: Optional. '2x2' or '5x5'.
    -> FITS image file.
    """
    if field not in {'elais-s1', 'cdfs'}:
        raise ValueError('field must be either "elais-s1" or "cdfs".')

    if wavelength not in {'ir', 'radio'}:
        raise ValueError('wavelength must be either "ir" or "radio".')

    cid = subject['metadata']['source']
    filename = '{}_{}.fits'.format(cid, wavelength)
    path = os.path.join(config['data_sources']['{}_fits'.format(field)], size,
                        filename)
    
    return astropy.io.fits.open(path, ignore_blank=True)


@require_atlas
def get_ir(subject, size='2x2'):
    """Returns the IR image of a subject.

    subject: RGZ subject dict, from the ATLAS survey.
    size: Optional. '2x2' or '5x5'.
    -> NumPy array.
    """
    if subject['metadata']['source'].startswith('C'):
        field = 'cdfs'
    else:
        field = 'elais-s1'

    with open_fits(subject, field, 'ir', size=size) as fits_file:
        return fits_file[0].data


@require_atlas
def get_ir_fits(subject, size='2x2'):
    """Returns the IR image of a subject as a FITS image.

    subject: RGZ subject dict, from the ATLAS survey.
    size: Optional. '2x2' or '5x5'.
    -> FITS image.
    """
    if subject['metadata']['source'].startswith('C'):
        field = 'cdfs'
    else:
        field = 'elais-s1'

    with open_fits(subject, field, 'ir', size=size) as fits_file:
        return fits_file[0]


@require_atlas
def get_radio(subject, size='2x2'):
    """Returns the radio image of a subject.

    subject: RGZ subject dict, from the ATLAS survey.
    size: Optional. '2x2' or '5x5'.
    -> NumPy array.
    """
    if subject['metadata']['source'].startswith('C'):
        field = 'cdfs'
    else:
        field = 'elais-s1'

    with open_fits(subject, field, 'radio', size=size) as fits_file:
        return fits_file[0].data


@require_atlas
def get_radio_fits(subject, size='2x2'):
    """Returns the radio image of a subject as a FITS image.

    subject: RGZ subject dict, from the ATLAS survey.
    size: Optional. '2x2' or '5x5'.
    -> FITS image.
    """
    if subject['metadata']['source'].startswith('C'):
        field = 'cdfs'
    else:
        field = 'elais-s1'

    with open_fits(subject, field, 'radio', size=size) as fits_file:
        return fits_file[0]


def get_contours(subject):
    """Fetches the radio contours of a subject.

    subject: RGZ subject dict.
    -> JSON dict.
    """
    # TODO(MatthewJA): Cache these.
    return requests.get(subject['location']['contours']).json()


@require_atlas
def get_potential_hosts(subject, cache_name, convert_to_px=True):
    """Finds the potential hosts for a subject.

    subject: RGZ subject dict.
    cache_name: Name of Gator cache.
    convert_to_px: Whether to convert coordinates to pixels. Default True; if
        False then coordinates will be RA/DEC.
    -> dict mapping (x, y) tuples to
        - flux at 3.6μm for aperture #2
        - flux at 4.5μm for aperture #2
        - flux at 5.8μm for aperture #2
        - flux at 8.0μm for aperture #2
        - flux at 24μm for aperture #2
        - stellarity index at 3.6μm
        - uncertainty in RA
        - uncertainty in DEC
    """

    if subject['metadata']['source'].startswith('C'):
        # CDFS
        catalog = 'chandra_cat_f05'
    else:
        # ELAIS-S1
        catalog = 'elaiss1_cat_f05'
    
    query = {
        'catalog': catalog,
        'spatial': 'box',
        'objstr': '{} {}'.format(*subject['coords']),
        'size': '120',
        'outfmt': '3',
    }
    url = 'http://irsa.ipac.caltech.edu/cgi-bin/Gator/nph-query'

    r = requests.get(url, params=query)
    votable = astropy.io.votable.parse_single_table(io.BytesIO(r.content),
                                                    pedantic=False)
    
    ras = votable.array['ra']
    decs = votable.array['dec']

    if convert_to_px:
        # Convert to px.
        fits = get_ir_fits(subject)
        wcs = astropy.wcs.WCS(fits.header)
        xs, ys = wcs.all_world2pix(ras, decs, 0)
    else:
        xs, ys = ras, decs
    
    # Get the astronomical features.
    out = {}  # Maps (x, y) to astronomical features.
    for x, y, row_idx in zip(xs, ys, range(votable.nrows)):
        row = votable.array[row_idx]
        out[x, y] = {
            'name': row['object'],
            'clon': row['clon'],
            'clat': row['clat'],
            'flux_ap2_36': row['flux_ap2_36'],
            'flux_ap2_45': row['flux_ap2_45'],
            'flux_ap2_58': row['flux_ap2_58'],
            'flux_ap2_80': row['flux_ap2_80'],
            'flux_ap2_24': row['flux_ap2_24'],
            'stell_36': row['stell_36'],
            'unc_ra': row['unc_ra'],
            'unc_dec': row['unc_dec'],
        }

    return out



import argparse
import logging

import h5py
import matplotlib
import matplotlib.pyplot as plot
import numpy
import sklearn.externals
import sklearn.metrics


font = {'family' : 'Palatino Linotype',
        'size'   : 22}

matplotlib.rc('font', **font)


def main(inputs_h5, training_h5, classifier_path, astro_transformer_path,
         image_transformer_path, use_astro=True, use_cnn=True, simple=False):
    classifier = sklearn.externals.joblib.load(classifier_path)
    astro_transformer = sklearn.externals.joblib.load(astro_transformer_path)
    image_transformer = sklearn.externals.joblib.load(image_transformer_path)

    testing_indices = inputs_h5['/atlas/cdfs/testing_indices'].value
    all_astro_inputs = astro_transformer.transform(training_h5['astro'].value)
    all_cnn_inputs = image_transformer.transform(
            training_h5['cnn_outputs'].value)
    all_inputs = numpy.hstack([all_astro_inputs, all_cnn_inputs])
    all_labels = training_h5['labels'].value
    
    inputs = all_inputs[testing_indices]
    labels = all_labels[testing_indices]
    probs = classifier.predict_proba(inputs)
    outputs = classifier.predict(inputs)
    
    precision, recall, _ = sklearn.metrics.precision_recall_curve(labels,
                                                                  probs[:, 1])
    plot.figure(figsize=(20, 10))
    plot.subplot(1, 2, 1)
    plot.plot(recall, precision)
    plot.xlabel('Recall')
    plot.ylabel('Precision')

    plot.subplot(1, 2, 2)
    fpr, tpr, _ = sklearn.metrics.roc_curve(labels, probs[:, 1])
    plot.plot(fpr, tpr)
    plot.xlabel('False positive rate')
    plot.ylabel('True positive rate')
    plot.show()

    cm = sklearn.metrics.confusion_matrix(labels, outputs)
    cm = cm.astype('float') / cm.sum(axis=1)[:, numpy.newaxis]
    plot.imshow(cm, interpolation='nearest', cmap=plot.cm.Blues)
    plot.colorbar()
    tick_marks = numpy.arange(2)
    plot.xticks(tick_marks, ['not host', 'host'], rotation=45)
    plot.yticks(tick_marks, ['not host', 'host'])
    plot.tight_layout()
    plot.ylabel('True label')
    plot.xlabel('Predicted label')
    plot.show()
    
    print('Accuracy: {:.02%}'.format(classifier.score(inputs, labels)))

    n_true_pos = numpy.sum(numpy.logical_and(labels == outputs, labels == 1))
    n_true_neg = numpy.sum(numpy.logical_and(labels == outputs, labels == 0))
    n_false_pos = numpy.sum(numpy.logical_and(labels != outputs, labels == 0))
    n_false_neg = numpy.sum(numpy.logical_and(labels != outputs, labels == 1))
    print('Balanced accuracy: {:.02%}'.format(0.5 * n_true_pos / (n_true_pos + n_false_neg) + 0.5 * n_true_neg / (n_true_neg + n_false_pos)))


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--inputs', default='crowdastro.h5',
                        help='HDF5 inputs data file')
    parser.add_argument('--training', default='training.h5',
                        help='HDF5 training data file')
    parser.add_argument('--classifier', default='classifier.pkl',
                        help='classifier file')
    parser.add_argument('--astro_transformer', default='astro_transformer.pkl',
                        help='astro transformer file')
    parser.add_argument('--image_transformer', default='image_transformer.pkl',
                        help='image transformer file')
    parser.add_argument('--no_astro', action='store_false', default=True,
                        help='ignore astro features')
    parser.add_argument('--no_cnn', action='store_false', default=True,
                        help='ignore CNN features')
    parser.add_argument('--simple', action='store_true', default=False,
                        help='use only single-AGN subjects')
    args = parser.parse_args()

    logging.root.setLevel(logging.DEBUG)

    with h5py.File(args.training, 'r') as training_h5:
        with h5py.File(args.inputs, 'r') as inputs_h5:
            main(inputs_h5, training_h5, args.classifier,
                 args.astro_transformer, args.image_transformer,
                 use_astro=args.no_astro, use_cnn=args.no_cnn,
                 simple=args.simple)



"""Tests classifiers on radio subjects.

Matthew Alger
The Australian National University
2016
"""

import argparse
import logging

import h5py
import numpy
import sklearn.externals.joblib
import sklearn.neighbors

from .config import config

PATCH_RADIUS = config['patch_radius']
PATCH_DIAMETER = PATCH_RADIUS * 2
ARCMIN = 1 / 60
FITS_HEIGHT = config['surveys']['atlas']['fits_height']
FITS_WIDTH = config['surveys']['atlas']['fits_height']
IMAGE_SIZE = FITS_HEIGHT * FITS_WIDTH


def test(inputs_h5, training_h5, classifier_path, astro_transformer_path,
         image_transformer_path, use_astro=True, use_cnn=True):
    """Tests classifiers.

    inputs_h5: Inputs HDF5 file.
    training_h5: Training HDF5 file.
    classifier_path: Path to classifier.
    astro_transformer_path: Path to astro transformer.
    image_transformer_path: Path to image transformer.
    """
    # TODO(MatthewJA): Figure out a neat way to test on subjects stored in
    # training_h5.
    classifier = sklearn.externals.joblib.load(classifier_path)
    astro_transformer = sklearn.externals.joblib.load(astro_transformer_path)
    image_transformer = sklearn.externals.joblib.load(image_transformer_path)

    assert training_h5.attrs['ir_survey'] == inputs_h5.attrs['ir_survey']
    n_static = 6 if training_h5.attrs['ir_survey'] == 'swire' else 5

    test_indices = training_h5['is_atlas_test'].value
    numeric_subjects = inputs_h5['/atlas/cdfs/numeric'][test_indices, :]

    n_correct = 0
    n_total = 0
    for subject in numeric_subjects:
        swire = subject[2 + IMAGE_SIZE:]
        nearby = swire < ARCMIN
        astro_inputs = numpy.minimum(training_h5['features'][nearby, :n_static],
                                     1500)
        image_inputs = training_h5['features'][nearby, n_static:]

        features = []
        if use_astro:
            features.append(astro_transformer.fit_transform(astro_inputs))
        if use_cnn:
            features.append(image_transformer.fit_transform(image_inputs))
        inputs = numpy.hstack(features)

        outputs = training_h5['labels'][nearby]

        selection = classifier.predict_proba(inputs)[:, 1].argmax()
        n_correct += outputs[selection]
        n_total += 1

    print('{:.02%}'.format(n_correct / n_total))


def _populate_parser(parser):
    parser.description = 'Tests classifiers on radio subjects.'
    parser.add_argument('--inputs', default='data/crowdastro.h5',
                        help='HDF5 crowdastro data file')
    parser.add_argument('--training', default='data/training.h5',
                        help='HDF5 training data file')
    parser.add_argument('--classifier', default='data/classifier.pkl',
                        help='classifier file')
    parser.add_argument('--astro_transformer',
                        default='data/astro_transformer.pkl',
                        help='astro transformer file')
    parser.add_argument('--image_transformer',
                        default='data/image_transformer.pkl',
                        help='image transformer file')
    parser.add_argument('--no_astro', action='store_false', default=True,
                        help='ignore astro features')
    parser.add_argument('--no_cnn', action='store_false', default=True,
                        help='ignore CNN features')


def _main(args):
    with h5py.File(args.training, 'r') as training_h5:
        with h5py.File(args.inputs, 'r') as inputs_h5:
            assert inputs_h5.attrs['version'] == '0.5.0'
            test(inputs_h5, training_h5, args.classifier,
                 args.astro_transformer, args.image_transformer,
                 use_astro=args.no_astro, use_cnn=args.no_cnn)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)



"""Tests classifiers on subjects.

Matthew Alger
The Australian National University
2016
"""

import argparse
import csv
import logging

import astropy.io
import astropy.wcs
import h5py
import matplotlib
import matplotlib.colors as co
import matplotlib.pyplot as plot
import numpy
import sklearn.externals.joblib
import sklearn.neighbors

from .config import config
from .data import get_ir, get_subject
from .show import contours

font = {'family' : 'Palatino Linotype',
        'size'   : 22}

matplotlib.rc('font', **font)

PATCH_RADIUS = config['patch_radius']
PATCH_DIAMETER = PATCH_RADIUS * 2
ARCMIN = 1 / 60

def test(inputs_h5, inputs_csv, training_h5, classifier_path,
         astro_transformer_path, image_transformer_path, use_astro=True,
         use_cnn=True):
    classifier = sklearn.externals.joblib.load(classifier_path)
    astro_transformer = sklearn.externals.joblib.load(astro_transformer_path)
    image_transformer = sklearn.externals.joblib.load(image_transformer_path)

    testing_indices = inputs_h5['/atlas/cdfs/testing_indices'].value
    swire_positions = inputs_h5['/swire/cdfs/catalogue'][:, :2]
    atlas_positions = inputs_h5['/atlas/cdfs/positions'].value
    all_astro_inputs = training_h5['astro'].value
    all_cnn_inputs = training_h5['cnn_outputs'].value
    all_labels = training_h5['labels'].value

    atlas_counts = {}  # ATLAS ID to number of objects in that subject.
    for consensus in inputs_h5['/atlas/cdfs/consensus_objects']:
        atlas_id = int(consensus[0])
        atlas_counts[atlas_id] = atlas_counts.get(atlas_id, 0) + 1

    simple_indices = []
    for atlas_id, count in atlas_counts.items():
        if count == 1 and atlas_id in testing_indices:
            simple_indices.append(atlas_id)
    print(simple_indices)

    atlas_positions = inputs_h5['/atlas/cdfs/positions']
    csvdicts = list(csv.DictReader(inputs_csv))
    for index, position in enumerate(atlas_positions[:100]):
        if index not in testing_indices or index not in simple_indices:
            continue

        swire_positions = training_h5['positions']
        swire_tree = sklearn.neighbors.KDTree(swire_positions)
        neighbours, distances = swire_tree.query_radius([position], ARCMIN,
                                                        return_distance=True)
        neighbours = neighbours[0]
        distances = distances[0]
        nearest_positions = swire_positions.value[neighbours]
        astro_inputs = all_astro_inputs[neighbours]
        astro_inputs[:, -1] = distances
        cnn_inputs = all_cnn_inputs[neighbours]
        labels = all_labels[neighbours]

        astro_inputs = astro_transformer.transform(astro_inputs)
        cnn_inputs = image_transformer.transform(cnn_inputs)

        probs = classifier.predict_proba(numpy.hstack([astro_inputs, cnn_inputs]))
    
        for row in csvdicts:
            if int(row['index']) == index and row['survey'] == 'atlas':
                zid = row['zooniverse_id']
                header = row['header']
                break
        subject = get_subject(zid)
        ir = get_ir(subject)
        wcs = astropy.wcs.WCS(astropy.io.fits.Header.fromstring(header))
        points = wcs.all_world2pix(nearest_positions, 1)
        print(zid)
        plot.figure(figsize=(20, 10))

        plot.subplot(1, 2, 1)
        plot.imshow(ir, cmap='gray', norm=co.LogNorm(vmin=ir.min(), vmax=ir.max()))
        contours(subject)
        plot.xlim((0, 200))
        plot.ylim((0, 200))
        plot.axis('off')
        plot.scatter(points[:, 0] - 151, points[:, 1] - 151, zorder=100, c=probs[:, 1], cmap='cool', s=100)

        plot.subplot(1, 2, 2)
        plot.scatter(range(len(probs)), sorted(probs[:, 1]), c=sorted(probs[:, 1]), cmap='cool', linewidth=0, s=100)
        plot.xlim((0, len(probs)))
        plot.ylim((0, 1))
        plot.xlabel('SWIRE object index')
        plot.ylabel('Classifier probability')
        plot.show()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--inputs', default='crowdastro.h5',
                        help='HDF5 inputs data file')
    parser.add_argument('--csv', default='crowdastro.csv',
                        help='CSV inputs data file')
    parser.add_argument('--training', default='training.h5',
                        help='HDF5 training data file')
    parser.add_argument('--classifier', default='classifier.pkl',
                        help='classifier file')
    parser.add_argument('--astro_transformer', default='astro_transformer.pkl',
                        help='astro transformer file')
    parser.add_argument('--image_transformer', default='image_transformer.pkl',
                        help='image transformer file')
    parser.add_argument('--no_astro', action='store_false', default=True,
                        help='ignore astro features')
    parser.add_argument('--no_cnn', action='store_false', default=True,
                        help='ignore CNN features')
    args = parser.parse_args()

    logging.root.setLevel(logging.DEBUG)

    with h5py.File(args.training, 'r') as training_h5:
        with h5py.File(args.inputs, 'r') as inputs_h5:
            with open(args.csv, 'r') as inputs_csv:
                test(inputs_h5, inputs_csv, training_h5, args.classifier,
                     args.astro_transformer, args.image_transformer,
                     use_astro=args.no_astro, use_cnn=args.no_cnn)



"""Finds the consensus crowd classifications in Radio Galaxy Zoo.

Matthew Alger
The Australian National University
2016
"""

import argparse
import collections
import csv
import itertools
import logging

import h5py
import numpy
import scipy.linalg.basic
import scipy.ndimage.filters
import scipy.ndimage.morphology
import scipy.stats
import sklearn.mixture
import sklearn.neighbors


MAX_RADIO_SIGNATURE_LENGTH = 50  # max number of components * individual
                                 # component signature size.


def maybe_mean(arr):
    arr = arr[~numpy.all(numpy.isnan(arr), axis=1)]
    if arr.shape[0] == 0:
        return (float('nan'), float('nan'))

    return arr.mean(axis=0)


def pg_means(points, significance=0.01, projections=24):
    """Find a consensus location with the PG-means algorithm.

    points: Array of points with dimension (N, 2).
    significance: Optional. Significance level for increasing the Gaussian
        count.
    projections: Optional. How many projections to try before accepting.
    -> (x, y), boolean of whether PG-means succeeded.
    """
    k = 1

    last_gmm = None
    while True:
        # Fit a Gaussian mixture model with k components.
        gmm = sklearn.mixture.GMM(n_components=k, covariance_type='full')
        try:
            gmm.fit(points)
        except ValueError:
            if last_gmm is None:
                return maybe_mean(points), False

            return last_gmm.means_[last_gmm.weights_.argmax()], False
        last_gmm = gmm

        for _ in range(projections):
            # Project the data to one dimension.
            projection_vector = numpy.random.random(size=(2,))
            projected_points = numpy.dot(points, projection_vector)
            # Project the model to one dimension. We need the CDF in one
            # dimension, so we'll sample some data points and project them.
            n_samples = 1000
            samples = numpy.dot(gmm.sample(n_samples), projection_vector)
            samples.sort()

            def cdf(x):
                for sample, y in zip(samples,
                                     numpy.arange(n_samples) / n_samples):
                    if sample >= x:
                        break
                return y

            _, p_value = scipy.stats.kstest(projected_points,
                                            numpy.vectorize(cdf))
            if p_value < significance:
                # Reject the null hypothesis.
                break
        else:
            # Null hypothesis was not broken.
            return gmm.means_[gmm.weights_.argmax()], True

        k += 1


def lowest_bic_gmm(points, min_k=1, max_k=5):
    """Find a consensus location by fitting a GMM with lowest BIC.

    points: Array of points with dimension (N, 2).
    min_k: Minimum number of components.
    max_k: Maximum number of components.
    -> (x, y), boolean of whether the fit succeeded
    """
    min_bic = float('inf')
    min_gmm = None
    for k in range(min_k, max_k):
        gmm = sklearn.mixture.GMM(n_components=k, covariance_type='full')
        try:
            gmm.fit(points)
        except ValueError:
            break
        bic = gmm.bic(points)
        if bic < min_bic:
            min_bic = bic
            min_gmm = gmm
    
    if not min_gmm:
        return points.mean(axis=0), False
    
    if sum(w == max(min_gmm.weights_) for w in min_gmm.weights_) > 1:
        success = False
    else:
        success = True
    
    return min_gmm.means_[min_gmm.weights_.argmax()], success


def kde(points):
    """Find a consensus location with the KDE algorithm.

    points: [[x, y]] NumPy array.
    -> (x, y) consensus location, boolean of whether KDE succeeded
    """
    if len(points) == 1:
        return points.mean(axis=0), False

    if len(points) == 0:
        return (float('nan'), float('nan')), False

    X, Y = numpy.mgrid[0:200, 0:200]
    positions = numpy.vstack([X.ravel(), Y.ravel()])
    try:
        kernel = scipy.stats.gaussian_kde(points.T)
    except scipy.linalg.basic.LinAlgError:
        logging.warning('LinAlgError in KD estimation.')
        return maybe_mean(points), False
    except ValueError as e:
        logging.warning('ValueError in KD estimation: %s', e.message)
        return maybe_mean(points), False

    kp = kernel(positions)
    if numpy.isnan(kp).sum() > 0:
        logging.warning('NaN in KD estimation.')
        return maybe_mean(points), False

    neighborhood = numpy.ones((10, 10))
    Z = numpy.reshape(kp.T, X.shape)
    local_max = scipy.ndimage.filters.maximum_filter(
            Z, footprint = neighborhood) == Z
    background = (Z == 0)
    eroded_background = scipy.ndimage.morphology.binary_erosion(
            background, structure = neighborhood, border_value=1)
    detected_peaks = local_max ^ eroded_background

    x_peak = float(X[Z == Z.max()][0])
    y_peak = float(Y[Z == Z.max()][0])
    return (x_peak, y_peak), True


def find_consensuses(f_h5, ir_survey):
    """Find Radio Galaxy Zoo crowd consensuses.

    f_h5: crowdastro HDF5 file.
    ir_survey: SWIRE or WISE.
    """
    if 'consensus_objects' in f_h5['/atlas/cdfs/']:
        del f_h5['/atlas/cdfs/consensus_objects']

    class_positions = f_h5['/atlas/cdfs/classification_positions']
    class_combinations = f_h5['/atlas/cdfs/classification_combinations']
    assert len(class_positions) == len(class_combinations)

    logging.debug('Finding consensuses for %d classifications.',
                  len(class_combinations))

    # Pre-build the IR tree.
    ir_coords = f_h5['/{}/cdfs/numeric'.format(ir_survey)][:, :2]
    ir_tree = sklearn.neighbors.KDTree(ir_coords)

    cons_positions = []
    cons_combinations = []

    # Data integrity and assumptions checks.
    assert numpy.array_equal(class_positions[:, 0],
                             class_combinations['index'])
    assert numpy.array_equal(class_positions[:, 0],
                             sorted(class_positions[:, 0]))

    pos_groups = itertools.groupby(class_positions, key=lambda z: z[0])
    com_groups = itertools.groupby(class_combinations, key=lambda z: z['index'])

    for (i, pos_group), (j, com_group) in zip(pos_groups, com_groups):
        assert i == j

        com_group = list(com_group)  # For multiple iterations.
        pos_group = list(pos_group)
        total_classifications = 0

        # Find the radio consensus. Be wary when counting: If there are multiple
        # AGNs identified in one subject, *that classification will appear
        # multiple times*. I'm going to deal with this by dividing the weight of
        # each classification by how many pipes it contains plus one.
        radio_counts = {}  # Radio signature -> Count
        for _, full_com, _ in com_group:
            count = radio_counts.get(full_com, 0)
            count += 1 / (full_com.count(b'|') + 1)
            radio_counts[full_com] = count

            total_classifications += 1 / (full_com.count(b'|') + 1)

        for count in radio_counts.values():
            # Despite the divisions, we should end up with integers overall.
            assert numpy.isclose(round(count), count)
        assert numpy.isclose(round(total_classifications),
                             total_classifications)

        radio_consensus = max(radio_counts, key=radio_counts.get)

        # Find the location consensus. For each radio combination, run a
        # location consensus function on the positions associated with that
        # combination.
        for radio_signature in radio_consensus.split(b'|'):
            percentage_consensus = (radio_counts[radio_consensus] /
                                    total_classifications)
            locations = []
            for (_, x, y), (_, full, radio) in zip(pos_group, com_group):
                if full == radio_consensus and radio == radio_signature:
                    locations.append((x, y))
            locations = numpy.array(locations)
            locations = locations[~numpy.all(numpy.isnan(locations), axis=1)]
            (x, y), success = lowest_bic_gmm(locations)

            if numpy.isnan(x) or numpy.isnan(y):
                logging.debug('Skipping NaN PG-means output.')
                continue

            # Match the (x, y) position to an IR object.
            dist, ind = ir_tree.query([(x, y)])

            # TODO(MatthewJA): Cut-off based on distance.
            # Since IR data is sorted, we can deal directly with indices.
            cons_positions.append((i, ind[0][0], success))
            cons_combinations.append((i, radio_signature, percentage_consensus))

    logging.debug('Found %d consensuses (before duplicate removal).',
                  len(cons_positions))

    # Remove duplicates. For training data, I don't really care if radio
    # combinations overlap (though I need to care if I generate a catalogue!) so
    # just take duplicated locations and pick the one with the highest radio
    # consensus that has success.
    cons_objects = {}  # Maps IR index to (ATLAS index, success,
                       #                   percentage_consensus)
    for (atlas_i, ir_i, success), (atlas_j, radio, percentage) in zip(
            cons_positions, cons_combinations):
        assert atlas_i == atlas_j

        if ir_i not in cons_objects:
            cons_objects[ir_i] = (atlas_i, success, percentage)
            continue

        if cons_objects[ir_i][1] and not success:
            # Preference successful KDE/PG-means.
            continue

        if not cons_objects[ir_i][1] and success:
            # Preference successful KDE/PG-means.
            cons_objects[ir_i] = (atlas_i, success, percentage)
            continue

        # If we get this far, we have the same success state. Choose based on
        # radio consensus.
        if percentage > cons_objects[ir_i][2]:
            cons_objects[ir_i] = (atlas_i, success, percentage)
            continue

    logging.debug('Found %d consensuses.', int(len(cons_objects)))

    cons_objects = numpy.array([(atlas_i, ir_i, success, percentage)
            for ir_i, (atlas_i, success, percentage)
            in sorted(cons_objects.items())])

    # Write rows to file.
    cons_objects = numpy.array(cons_objects, dtype=float)
    f_h5['/atlas/cdfs/'].create_dataset('consensus_objects',
                                        data=cons_objects, dtype=float)


def _populate_parser(parser):
    parser.description = 'Generates Radio Galaxy Zoo crowd consensus ' \
                         'classifications.'
    parser.add_argument('--h5', default='data/crowdastro.h5',
                        help='HDF5 IO file')


def _main(args):
    with h5py.File(args.h5, 'r+') as f_h5:
        assert f_h5.attrs['version'] == '0.5.1'
        ir_survey = f_h5.attrs['ir_survey']
        find_consensuses(f_h5, ir_survey)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    _populate_parser(parser)
    args = parser.parse_args()
    _main(args)
